"""Module that contain iterator used for dynamic data."""

import torch
from itertools import cycle
from eole.constants import CorpusTask
from eole.inputters.text_corpus import get_corpora, build_corpora_iters
from eole.inputters.text_utils import (
    text_sort_key,
    transform_bucket,
    numericalize,
    tensorify,
)
from eole.config.data import Dataset
from eole.utils.logging import init_logger, logger
from eole.utils.misc import RandomShuffler, get_device
from torch.utils.data import DataLoader


class MixingStrategy(object):
    """Mixing strategy that should be used in Data Iterator."""

    def __init__(self, iterables, weights):
        """Initilize neccessary attr."""
        self._valid_iterable(iterables, weights)
        self.iterables = iterables
        self.weights = weights

    def _valid_iterable(self, iterables, weights):
        iter_keys = iterables.keys()
        weight_keys = weights.keys()
        if iter_keys != weight_keys:
            raise ValueError(f"keys in {iterables} & {weights} should be equal.")

    def __iter__(self):
        raise NotImplementedError


class SequentialMixer(MixingStrategy):
    """Generate data sequentially from `iterables` which is exhaustible."""

    def _iter_datasets(self):
        for ds_name, ds_weight in self.weights.items():
            for _ in range(ds_weight):
                yield ds_name

    def __iter__(self):
        for ds_name in self._iter_datasets():
            iterable = self.iterables[ds_name]
            yield from iterable


class WeightedMixer(MixingStrategy):
    """A mixing strategy that mix data weightedly and iterate infinitely."""

    def __init__(self, iterables, weights):
        super().__init__(iterables, weights)
        self._iterators = {}
        self._counts = {}
        for ds_name in self.iterables.keys():
            self._reset_iter(ds_name)

    def _logging(self):
        """Report corpora loading statistics."""
        msgs = []
        # patch to log stdout spawned processes of dataloader
        logger = init_logger()
        for ds_name, ds_count in self._counts.items():
            msgs.append(f"\t\t\t* {ds_name}: {ds_count}")
        logger.info("Weighted corpora loaded so far:\n" + "\n".join(msgs))

    def _reset_iter(self, ds_name):
        self._iterators[ds_name] = iter(self.iterables[ds_name])
        self._counts[ds_name] = self._counts.get(ds_name, 0) + 1
        self._logging()

    def _iter_datasets(self):
        for ds_name, ds_weight in self.weights.items():
            for _ in range(ds_weight):
                yield ds_name

    def __iter__(self):
        for ds_name in cycle(self._iter_datasets()):
            iterator = self._iterators[ds_name]
            try:
                item = next(iterator)
            except StopIteration:
                self._reset_iter(ds_name)
                iterator = self._iterators[ds_name]
                item = next(iterator)
            finally:
                yield item


class DynamicDatasetIter(torch.utils.data.IterableDataset):
    """Yield batch from (multiple) plain text corpus.

    Args:
        corpora (dict[str, ParallelCorpus]): collections of corpora to iterate;
        corpora_info (dict[str, dict]): corpora infos correspond to corpora;
        transforms (dict[str, Transform]): transforms may be used by corpora;
        vocabs (dict[str, Vocab]): vocab dict for convert corpora into Tensor;
        task (str): CorpusTask.TRAIN/VALID/INFER;
        batch_type (str): batching type to count on, choices=[tokens, sents];
        batch_size (int): numbers of examples in a batch;
        batch_size_multiple (int): make batch size multiply of this;
        data_type (str): input data type, currently only text;
        bucket_size (int): accum this number of examples in a dynamic dataset;
        bucket_size_init (int): initialize the bucket with this
        amount of examples;
        bucket_size_increment (int): increment the bucket
        size with this amount of examples;
        skip_empty_level (str): security level when encouter empty line;
        stride (int): iterate data files with this stride;
        offset (int): iterate data files with this offset.

    Attributes:
        sort_key (function): functions define how to sort examples;
        mixer (MixingStrategy): the strategy to iterate corpora.
    """

    def __init__(
        self,
        corpora,
        corpora_info,
        transforms,
        vocabs,
        task,
        batch_type,
        batch_size,
        batch_size_multiple,
        data_type="text",
        bucket_size=2048,
        bucket_size_init=-1,
        bucket_size_increment=0,
        device=torch.device("cpu"),
        skip_empty_level="warning",
        stride=1,
        offset=0,
        score_threshold=0,
        left_pad=False,
        model_type=None,
        image_patch_size=16,
        image_size=1024,
        adapter="llava",
    ):
        super(DynamicDatasetIter).__init__()
        self.corpora = corpora
        self.transforms = transforms
        self.vocabs = vocabs
        self.corpora_info = corpora_info
        self.task = task
        self.init_iterators = False
        self.batch_size = batch_size
        self.batch_type = batch_type
        self.batch_size_multiple = batch_size_multiple
        self.sort_key = text_sort_key
        self.bucket_size = bucket_size
        self.bucket_size_init = bucket_size_init
        self.bucket_size_increment = bucket_size_increment
        self.device = device
        if stride <= 0:
            raise ValueError(f"Invalid argument for stride={stride}.")
        self.stride = stride
        self.offset = offset
        self.score_threshold = score_threshold
        if skip_empty_level not in ["silent", "warning", "error"]:
            raise ValueError(f"Invalid argument skip_empty_level={skip_empty_level}")
        self.skip_empty_level = skip_empty_level
        self.random_shuffler = RandomShuffler()
        self.bucket_idx = 0
        # TODO: we might want to enable some hybrid mode (default left_pad True for LM, else False)
        if task != CorpusTask.TRAIN and left_pad:
            self.left_pad = True
        else:
            self.left_pad = False
        self.model_type = model_type
        self.image_patch_size = image_patch_size
        self.image_size = image_size
        self.adapter = adapter

    @classmethod
    def from_config(
        cls,
        corpora,
        transforms,
        vocabs,
        config,
        task,
        device,
        stride=1,
        offset=0,
        model_type=None,
    ):
        """Initilize `DynamicDatasetIter` with options parsed from `opt`."""
        from eole.config.run import PredictConfig

        # patch for now, might do better later
        if isinstance(config, PredictConfig):
            running_config = config
        else:
            running_config = config.training

        corpora_info = {}
        batch_size = running_config.valid_batch_size if (task == CorpusTask.VALID) else running_config.batch_size
        # we might want to define proper infer/train classes instead of these conditions
        if task != CorpusTask.INFER:
            if running_config.batch_size_multiple is not None:
                batch_size_multiple = running_config.batch_size_multiple
            else:
                batch_size_multiple = 8 if running_config.compute_dtype == "fp16" else 1
            corpora_info = config.data
            bucket_size = running_config.bucket_size  # this should be in data rather than training?
            bucket_size_init = running_config.bucket_size_init
            bucket_size_increment = running_config.bucket_size_increment
            skip_empty_level = config.skip_empty_level
        else:
            batch_size_multiple = 1
            corpora_info[CorpusTask.INFER] = Dataset()
            corpora_info[CorpusTask.INFER].transforms = config.transforms
            corpora_info[CorpusTask.INFER].weight = 1
            # bucket_size = batch_size
            bucket_size = 16384
            bucket_size_init = -1
            bucket_size_increment = 0
            skip_empty_level = "warning"
        try:
            image_patch_size = config.model.encoder.patch_size * config.model.spatial_merge_size
        except AttributeError:
            image_patch_size = 16
        try:
            image_size = config.model.encoder.image_size
        except AttributeError:
            image_size = 0
        return cls(
            corpora,
            corpora_info,
            transforms,
            vocabs,
            task,
            running_config.batch_type,
            batch_size,
            batch_size_multiple,
            data_type=running_config.data_type,  # this is not super clear
            bucket_size=bucket_size,
            bucket_size_init=bucket_size_init,
            bucket_size_increment=bucket_size_increment,
            device=device,
            skip_empty_level=skip_empty_level,
            stride=stride,
            offset=offset,
            score_threshold=0 if isinstance(config, PredictConfig) else running_config.score_threshold,
            left_pad=getattr(config.model, "left_pad", False),
            model_type=model_type if model_type is not None else getattr(config.model, "model_type", None),
            image_patch_size=image_patch_size,
            image_size=image_size,
            adapter=getattr(config.model, "adapter", None),
        )

    def _init_datasets(self, worker_id):
        if self.num_workers > 0:
            stride = self.stride * self.num_workers
            offset = self.offset * self.num_workers + worker_id
        else:
            stride = self.stride
            offset = self.offset
        datasets_iterables = build_corpora_iters(
            self.corpora,
            self.transforms,
            self.corpora_info,
            skip_empty_level=self.skip_empty_level,
            stride=stride,
            offset=offset,
            image_patch_size=self.image_patch_size,
            image_size=self.image_size,
            adapter=self.adapter,
        )
        datasets_weights = {ds_name: int(self.corpora_info[ds_name].weight) for ds_name in datasets_iterables.keys()}
        if self.task == CorpusTask.TRAIN:
            self.mixer = WeightedMixer(datasets_iterables, datasets_weights)
        else:
            self.mixer = SequentialMixer(datasets_iterables, datasets_weights)
        self.init_iterators = True

    def _tuple_to_json_with_tokIDs(self, tuple_bucket):
        bucket = []
        tuple_bucket = transform_bucket(self.task, tuple_bucket, self.score_threshold)
        for example in tuple_bucket:
            if example is not None:
                bucket.append(numericalize(self.vocabs, example, model_type=self.model_type, task=self.task))

        return bucket

    def _add_indice(self, bucket):
        indice = 0
        indexed_bucket = []
        for ex in bucket:
            indexed_bucket.append((ex, indice))
            indice += 1
        return indexed_bucket

    def _bucketing(self):
        """
        Add up to bucket_size examples from the mixed corpora according
        to the above strategy. example tuple is converted to json and
        tokens numericalized.
        """
        bucket = []
        if self.bucket_size_init > 0:
            _bucket_size = self.bucket_size_init
        else:
            _bucket_size = self.bucket_size

        for ex in self.mixer:
            bucket.append(ex)
            if len(bucket) == _bucket_size:
                yield (self._tuple_to_json_with_tokIDs(bucket), self.bucket_idx)
                self.bucket_idx += 1
                bucket = []
                if _bucket_size < self.bucket_size:
                    _bucket_size += self.bucket_size_increment
                else:
                    _bucket_size = self.bucket_size
        if bucket:
            yield (self._tuple_to_json_with_tokIDs(bucket), self.bucket_idx)

    def batch_iter(self, data, batch_size, batch_type="sents", batch_size_multiple=1):
        """Yield elements from data in chunks of batch_size,
        where each chunk size is a multiple of batch_size_multiple.
        """

        def batch_size_fn(nbsents, maxlen):
            if batch_type == "sents":
                return nbsents
            elif batch_type == "tokens":
                return nbsents * maxlen
            else:
                raise ValueError(f"Invalid argument batch_type={batch_type}")

        def max_src_tgt(ex):
            """return the max tokens btw src and tgt in the sequence."""
            if ex.get("tgt", None) is not None:
                return max(len(ex["src"]["src_ids"]), len(ex["tgt"]["tgt_ids"]))
            return len(ex["src"]["src_ids"])

        minibatch, maxlen, size_so_far, seen = [], 0, 0, set()
        for ex, indice in data:
            src = ex["src"]["src"]
            if src not in seen or (self.task != CorpusTask.TRAIN):
                seen.add(src)
                minibatch.append((ex, indice))
                nbsents = len(minibatch)
                maxlen = max(max_src_tgt(ex), maxlen)
                size_so_far = batch_size_fn(nbsents, maxlen)
                if size_so_far >= batch_size:
                    overflowed = 1 if size_so_far > batch_size else 0
                    if batch_size_multiple > 1:
                        overflowed += (nbsents - overflowed) % batch_size_multiple
                    if overflowed == 0:
                        yield minibatch
                        minibatch, maxlen, size_so_far, seen = [], 0, 0, set()
                    else:
                        if overflowed == nbsents:
                            logger.warning(
                                "The batch will be filled until we reach"
                                " %d, its size may exceed %d tokens" % (batch_size_multiple, batch_size)
                            )
                        else:
                            yield minibatch[:-overflowed]
                            minibatch = minibatch[-overflowed:]
                            maxlen = max([max_src_tgt(ex) for ex, ind in minibatch])
                            size_so_far = batch_size_fn(len(minibatch), maxlen)
                            seen = set()

        if minibatch:
            yield minibatch

    def __iter__(self):
        for bucket, bucket_idx in self._bucketing():
            bucket = self._add_indice(bucket)
            bucket = sorted(bucket, key=lambda x: self.sort_key(x[0]))
            p_batch = list(
                self.batch_iter(
                    bucket,
                    self.batch_size,
                    batch_type=self.batch_type,
                    batch_size_multiple=self.batch_size_multiple,
                )
            )
            # For TRAIN we shuffle batches within the bucket
            # otherwise sequential
            if self.task == CorpusTask.TRAIN:
                p_batch = self.random_shuffler(p_batch)
            for i, minibatch in enumerate(p_batch):
                # for specific case of rnn_packed need to be sorted
                # within the batch
                if self.task == CorpusTask.TRAIN:
                    minibatch.sort(key=lambda x: self.sort_key(x[0]), reverse=True)
                tensor_batch = tensorify(self.vocabs, minibatch, self.device, self.left_pad)
                yield (tensor_batch, bucket_idx)


class OnDeviceDatasetIter:
    def __init__(self, data_iter, device):
        self.data_iter = data_iter
        self.device = device

    def __iter__(self):
        for tensor_batch, bucket_idx in self.data_iter:
            for key in tensor_batch.keys():
                if key not in [
                    "cid",
                    "ind_in_bucket",
                    "cid_line_number",
                    "left_pad",
                ]:
                    if isinstance(tensor_batch[key], list):
                        tensor_batch[key] = [t.to(self.device) for t in tensor_batch[key]]
                    elif tensor_batch[key] is not None:
                        tensor_batch[key] = tensor_batch[key].to(self.device)
            yield (tensor_batch, bucket_idx)


def build_dynamic_dataset_iter(
    config,
    transforms,
    vocabs,
    task=CorpusTask.TRAIN,
    stride=1,
    offset=0,
    src=None,
    tgt=None,
    align=None,
    device_id=-1,
    model_type=None,
):
    """
    Build `DynamicDatasetIter` from opt.
    if src, tgt,align are passed then dataset is built from those lists
    instead of opt.[src, tgt, align]
    Typically this function is called for CorpusTask.[TRAIN,VALID,INFER]
    from the main train / predict scripts
    We disable automatic batching in the DataLoader.
    The custom optimized batching is performed by the
    custom class DynamicDatasetIter inherited from IterableDataset
    (and not by a custom collate function).
    We load opt.bucket_size examples, sort them and yield
    mini-batchs of size opt.batch_size.
    The bucket_size must be large enough to ensure homogeneous batches.
    Each worker will load opt.prefetch_factor mini-batches in
    advance to avoid the GPU waiting during the refilling of the bucket.
    """
    corpora = get_corpora(config, task, src=src, tgt=tgt, align=align)
    if corpora is None:
        assert task != CorpusTask.TRAIN, "only valid corpus is ignorable."
        return None
    device = get_device(device_id=device_id) if device_id >= 0 else torch.device("cpu")
    if hasattr(config, "training"):
        num_workers = getattr(config.training, "num_workers", 0)
    else:
        num_workers = 0
    if num_workers == 0 or task == CorpusTask.INFER:
        # single thread - create batch directly on GPU if device is gpu
        data_iter = DynamicDatasetIter.from_config(
            corpora,
            transforms,
            vocabs,
            config,
            task,
            stride=stride,
            offset=offset,
            device=device,
            model_type=model_type,
        )
        data_iter.num_workers = num_workers
        data_iter._init_datasets(0)  # when workers=0 init_fn not called
        return data_iter
    else:
        # multithread faster to create batch on CPU in each thread and then move it to gpu
        data_iter = DynamicDatasetIter.from_config(
            corpora,
            transforms,
            vocabs,
            config,
            task,
            stride=stride,
            offset=offset,
            device=torch.device("cpu"),
        )
        data_iter.num_workers = num_workers
        data_loader = DataLoader(
            data_iter,
            batch_size=None,
            pin_memory=True,
            multiprocessing_context="spawn",
            num_workers=data_iter.num_workers,
            worker_init_fn=data_iter._init_datasets,
            prefetch_factor=config.training.prefetch_factor,
        )
        # Move tensor_batch from cpu to device
        return OnDeviceDatasetIter(data_loader, device)



import os
import importlib
import inspect


class BaseBin(object):
    """
    Abstract base class for tools.
    """

    @classmethod
    def add_args(cls, parser):
        raise NotImplementedError

    @classmethod
    def run(cls, parser):
        raise NotImplementedError


os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
os.environ["VLLM_LOGGING_LEVEL"] = "CRITICAL"
AVAILABLE_BINS = {}


def register_bin(name):
    """Wrapper to automatically register available tools."""

    def register_bin_cls(cls):
        # patch to ignore standalone imports to prevent conflicts (e.g. in build docs)
        if not (cls.__module__.startswith("eole")):
            return cls
        # bin_type is retrieved from bin parent dirname
        bin_type = os.path.basename(os.path.dirname(inspect.stack()[1].filename))
        if bin_type in AVAILABLE_BINS:
            if name in AVAILABLE_BINS[bin_type] and cls != AVAILABLE_BINS[bin_type][name]:
                raise ValueError(
                    f'Name "{name}" registered twice for different classes: \n'
                    f"{AVAILABLE_BINS[bin_type][name]} != {cls}"
                )
            else:
                AVAILABLE_BINS[bin_type][name] = cls
        else:
            AVAILABLE_BINS[bin_type] = {name: cls}
        return cls

    return register_bin_cls


# Auto import python files in all subdirectories
bin_dir = os.path.dirname(__file__)
for dir in os.listdir(bin_dir):
    abs_dir = os.path.join(os.path.dirname(__file__), dir)
    if os.path.isdir(abs_dir):
        for file in os.listdir(abs_dir):
            _path = os.path.join(abs_dir, file)
            if not file.startswith("_") and not file.startswith(".") and (file.endswith(".py") or os.path.isdir(_path)):
                file_name = file[: file.find(".py")] if file.endswith(".py") else file
                module = importlib.import_module(f"eole.bin.{dir}." + file_name)



import torch
from torch.nn.functional import log_softmax
from torch.nn.utils.rnn import pad_sequence
from itertools import count, zip_longest
from copy import deepcopy
from time import time
from math import exp
import codecs

from eole.transforms import TransformPipe, AVAILABLE_TRANSFORMS
from eole.constants import DefaultTokens
from eole.predict.prediction import PredictionBuilder
from eole.utils.misc import set_random_seed, report_matrix, sequence_mask, get_device
from eole.utils.alignment import build_align_pharaoh


class Inference(object):
    """Predict a batch of sentences with a saved model.

    Args:
        model (eole.modules.BaseModel): Model to use for prediction
        vocabs (dict[str, Vocab]): A dict
            mapping each side's Vocab.
        gpu (int): GPU device. Set to negative for no GPU.
        n_best (int): How many beams to wait for.
        min_length (int): See
            :class:`eole.predict.decode_strategy.DecodeStrategy`.
        max_length (int): See
            :class:`eole.predict.decode_strategy.DecodeStrategy`.
        beam_size (int): Number of beams.
        top_p (float): See
            :class:`eole.predict.greedy_search.GreedySearch`.
        top_k (int): See
            :class:`eole.predict.greedy_search.GreedySearch`.
        temperature (float): See
            :class:`eole.predict.greedy_search.GreedySearch`.
        stepwise_penalty (bool): Whether coverage penalty is applied every step
            or not.
        dump_beam (bool): Debugging option.
        block_ngram_repeat (int): See
            :class:`eole.predict.decode_strategy.DecodeStrategy`.
        ignore_when_blocking (set or frozenset): See
            :class:`eole.predict.decode_strategy.DecodeStrategy`.
        replace_unk (bool): Replace unknown token.
        tgt_file_prefix (bool): Force the predictions begin with provided -tgt.
        data_type (str): Source data type.
        verbose (bool): Print/log every prediction.
        report_time (bool): Print/log total time/frequency.
        global_scorer (eole.predict.GNMTGlobalScorer): Prediction
            scoring/reranking object.
        report_score (bool) : Whether to report scores
        logger (logging.Logger or NoneType): Logger.
    """

    def __init__(
        self,
        model,
        vocabs,
        gpu=-1,
        n_best=1,
        min_length=0,
        max_length=100,
        max_length_ratio=1.5,
        ratio=0.0,
        beam_size=30,
        top_k=0,
        top_p=0.0,
        temperature=1.0,
        stepwise_penalty=None,
        dump_beam=False,
        block_ngram_repeat=0,
        ignore_when_blocking=frozenset(),
        replace_unk=False,
        ban_unk_token=False,
        tgt_file_prefix=False,
        phrase_table="",
        data_type="text",
        verbose=False,
        report_time=False,
        global_scorer=None,
        report_align=False,
        gold_align=False,
        report_score=True,
        logger=None,
        seed=-1,
        with_score=False,
        estim_only=False,
        return_gold_log_probs=False,
        add_estimator=False,
        estimator_type="average",
        optional_eos=[],
        id_tokenization=False,
        image_token_id=10,
        fuse_kvq=False,
        fuse_gate=False,
    ):
        self.model = model
        if hasattr(self.model, "decoder") and self.model.decoder is not None:
            if hasattr(self.model.decoder, "transformer_layers"):
                for layer in self.model.decoder.transformer_layers:
                    if fuse_kvq:
                        layer.self_attn._fuse_KVQ()
                    if fuse_gate and getattr(layer.mlp, "up_proj", None) is not None:
                        layer.mlp._fuse_gate()
        if hasattr(self.model, "encoder") and self.model.encoder is not None:
            if hasattr(self.model.encoder, "transformer_layers"):
                for layer in self.model.encoder.transformer_layers:
                    if fuse_kvq:
                        layer.self_attn._fuse_KVQ()
                    if fuse_gate and getattr(layer.mlp, "up_proj", None) is not None:
                        layer.mlp._fuse_gate()

        self.vocabs = vocabs
        self._tgt_vocab = vocabs["tgt"]
        self._tgt_eos_idx = [vocabs["tgt"].lookup_token(vocabs.get("specials", {}).get("eos_token", ""))] + [
            vocabs["tgt"].lookup_token(eos_token) for eos_token in optional_eos
        ]
        # defaulting to DefaultTokens.PAD might not always work
        self._tgt_pad_idx = vocabs["tgt"].lookup_token(vocabs.get("specials", {}).get("pad_token", DefaultTokens.PAD))
        self._src_pad_idx = vocabs["src"].lookup_token(vocabs.get("specials", {}).get("pad_token", DefaultTokens.PAD))
        self._tgt_bos_idx = vocabs["tgt"].lookup_token(vocabs.get("specials", {}).get("bos_token", ""))
        self._tgt_unk_idx = vocabs["tgt"].lookup_token(vocabs.get("specials", {}).get("unk_token", ""))
        self._tgt_sep_idx = vocabs["tgt"].lookup_token(DefaultTokens.SEP)
        self._tgt_start_with = vocabs["tgt"].lookup_token(vocabs["decoder_start_token"])
        self._tgt_vocab_len = len(self._tgt_vocab)

        self._gpu = gpu
        self._use_gpu = gpu > -1
        self._dev = get_device(self._gpu) if self._use_gpu else torch.device("cpu")

        self.n_best = n_best
        self.max_length = max_length
        self.max_length_ratio = max_length_ratio

        self.beam_size = beam_size
        self.temperature = temperature
        self.top_k = top_k
        self.top_p = top_p

        self.min_length = min_length
        self.ban_unk_token = ban_unk_token
        self.ratio = ratio
        self.stepwise_penalty = stepwise_penalty
        self.dump_beam = dump_beam
        self.block_ngram_repeat = block_ngram_repeat
        self.ignore_when_blocking = ignore_when_blocking
        self._exclusion_idxs = {self._tgt_vocab[t] for t in self.ignore_when_blocking}
        self.replace_unk = replace_unk
        if self.replace_unk and not self.model.decoder.attentional:
            raise ValueError("replace_unk requires an attentional decoder.")
        self.tgt_file_prefix = tgt_file_prefix
        self.phrase_table = phrase_table
        self.data_type = data_type
        self.verbose = verbose
        self.report_time = report_time
        self.step0_time = 0.0

        self.global_scorer = global_scorer
        if self.global_scorer.has_cov_pen and not self.model.decoder.attentional:
            raise ValueError("Coverage penalty requires an attentional decoder.")
        self.report_align = report_align
        self.gold_align = gold_align
        self.report_score = report_score
        self.logger = logger

        self.use_filter_pred = False
        self._filter_pred = None

        # for debugging
        self.beam_trace = self.dump_beam != ""
        self.beam_accum = None
        if self.beam_trace:
            self.beam_accum = {
                "predicted_ids": [],
                "beam_parent_ids": [],
                "scores": [],
                "log_probs": [],
            }

        set_random_seed(seed, self._use_gpu)
        self.with_score = with_score
        self.estim_only = estim_only

        self.return_gold_log_probs = return_gold_log_probs
        self.add_estimator = add_estimator
        self.estimator_type = estimator_type
        self.id_tokenization = id_tokenization
        self.image_token_id = image_token_id

    @classmethod
    def from_config(
        cls,
        model,
        vocabs,
        config,  # running/predict config
        model_config,
        device_id=0,
        global_scorer=None,
        report_align=False,
        report_score=True,
        logger=None,
    ):
        """Alternate constructor.

        Args:
            model (eole.modules.BaseModel): See :func:`__init__()`.
            vocabs (dict[str, Vocab]): See
                :func:`__init__()`.
            opt (argparse.Namespace): Command line options
            model_opt (argparse.Namespace): Command line options saved with
                the model checkpoint.
            global_scorer (eole.predict.GNMTGlobalScorer): See
                :func:`__init__()`..
            report_align (bool) : See :func:`__init__()`.
            report_score (bool) : See :func:`__init__()`.
            logger (logging.Logger or NoneType): See :func:`__init__()`.
        """
        # TODO: maybe add dynamic part

        id_tokenization = False
        if hasattr(model_config, "encoder") and model_config.encoder is not None:
            image_token_id = getattr(model_config.encoder, "image_token_id", 10)
        else:
            image_token_id = 10
        if len(config.transforms) > 0:
            tail_transform_cls = AVAILABLE_TRANSFORMS.get(config.transforms[-1], None)
            if getattr(tail_transform_cls, "output_type", None) == "ids":
                id_tokenization = True

        return cls(
            model,
            vocabs,
            gpu=device_id,
            n_best=config.n_best,
            min_length=config.min_length,
            max_length=config.max_length,
            max_length_ratio=config.max_length_ratio,
            ratio=config.ratio,
            beam_size=config.beam_size,
            top_k=config.top_k,
            top_p=config.top_p,
            temperature=config.temperature,
            stepwise_penalty=config.stepwise_penalty,
            dump_beam=config.dump_beam,
            block_ngram_repeat=config.block_ngram_repeat,
            ignore_when_blocking=set(config.ignore_when_blocking),
            replace_unk=config.replace_unk,
            ban_unk_token=config.ban_unk_token,
            tgt_file_prefix=config.tgt_file_prefix,
            phrase_table=config.phrase_table,
            data_type=config.data_type,
            verbose=config.verbose,
            report_time=config.report_time,
            global_scorer=global_scorer,
            report_align=report_align,
            gold_align=config.gold_align,
            report_score=report_score,
            logger=logger,
            seed=config.seed,
            with_score=config.with_score,
            estim_only=config.estim_only,
            add_estimator=model_config.add_estimator,
            estimator_type=model_config.estimator_type,
            optional_eos=config.optional_eos,
            id_tokenization=id_tokenization,
            image_token_id=image_token_id,
            fuse_kvq=config.fuse_kvq,
            fuse_gate=config.fuse_gate,
        )

    def _log(self, msg):
        if self.logger:
            self.logger.info(msg)
        else:
            print(msg)

    def _gold_score(self, batch, enc_out, src_len, enc_final_hs, batch_size, src):
        if "tgt" in batch.keys() and not self.tgt_file_prefix:
            gs, glp = self._score_target(batch, enc_out, src_len)
            self.model.decoder.init_state(src=src, enc_out=enc_out, enc_final_hs=enc_final_hs)
        else:
            gs = [0] * batch_size
            glp = None
        return gs, glp

    def update_settings(self, **kwargs):
        # we probably would need some validation at some point
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)

    def _predict(
        self,
        infer_iter,
        transform=None,
        attn_debug=False,
        align_debug=False,
        phrase_table="",
    ):
        """Predict content of ``src`` and get gold scores from ``tgt``.

        Args:
            infer_iter: tensored batch iterator from DynamicDatasetIter
            attn_debug (bool): enables the attention logging
            align_debug (bool): enables the word alignment logging

        Returns:
            (`list`, `list`)

            * all_scores is a list of `batch_size` lists of `n_best` scores
            * all_predictions is a list of `batch_size` lists
                of `n_best` predictions
        """
        transform_pipe = TransformPipe.build_from([transform[name] for name in transform]) if transform else None
        prediction_builder = PredictionBuilder(
            self.vocabs,
            self.n_best,
            self.replace_unk,
            self.phrase_table,
            self._tgt_eos_idx,
            self.id_tokenization,
        )

        # Statistics
        counter = count(1)
        pred_score_total, estim_total, pred_words_total = 0, 0, 0
        gold_score_total, gold_words_total = 0, 0

        all_scores = []
        all_estim = []
        all_predictions = []

        start_time = time()
        self.step0_time = []

        def _maybe_retranslate(translations, batch):
            """Here we handle the cases of mismatch in number of segments
            between source and target. We re-translate seg by seg."""
            trans_copy = deepcopy(translations)
            for j, trans in enumerate(translations):
                if (trans.src == self._tgt_sep_idx).sum().item() != trans.pred_sents[0].count(DefaultTokens.SEP):
                    self._log("Mismatch in number of ((newline))")

                    # we rebuild a small batch made of the sub-segments
                    # in the long segment.
                    idx = (trans.src == self._tgt_sep_idx).nonzero()

                    sub_src = []
                    start_idx = 0
                    for i in range(len(idx)):
                        end_idx = idx[i]
                        sub_src.append(trans.src[start_idx:end_idx])
                        start_idx = end_idx + 1
                    end_idx = trans.src.ne(self._tgt_pad_idx).sum() - 1

                    sub_src.append(trans.src[start_idx:end_idx])
                    t_sub_src = pad_sequence(sub_src, batch_first=True, padding_value=self._tgt_pad_idx)
                    t_sub_src_len = t_sub_src.ne(self._tgt_pad_idx).sum(1)
                    t_sub_src_ind = [i for i in range(len(sub_src))]

                    device = batch["src"].device
                    t_sub_batch = {
                        "src": t_sub_src.to(device),
                        "srclen": t_sub_src_len.to(device),
                        "ind_in_bucket": t_sub_src_ind,
                    }
                    # new sub-batch ready to be predicted
                    sub_data = self.predict_batch(t_sub_batch, attn_debug)
                    sub_trans = prediction_builder.from_batch(sub_data)

                    # we re-insert the sub-batch in the initial predictions
                    nbest = len(sub_trans[0].pred_sents)
                    combined = [[] for _ in range(nbest)]

                    for sub_tr in sub_trans:
                        for i in range(nbest):
                            combined[i].extend(
                                [tok for tok in sub_tr.pred_sents[i] if tok != DefaultTokens.SEP] + [DefaultTokens.SEP]
                            )

                    trans_copy[j].pred_sents = [seq[:-1] for seq in combined]

            return trans_copy

        def _process_bucket(bucket_predictions):
            bucket_scores = []
            bucket_estims = []
            bucket_preds = []
            bucket_score = 0
            bucket_estim = 0
            bucket_words = 0
            bucket_gold_score = 0
            bucket_gold_words = 0
            voc_src = self.vocabs["src"].ids_to_tokens
            bucket_predictions = sorted(bucket_predictions, key=lambda x: x.ind_in_bucket)
            for trans in bucket_predictions:
                bucket_scores += [trans.pred_scores[: self.n_best]]
                if len(trans.pred_scores) > 0:
                    bucket_score += trans.pred_scores[0]
                bucket_estims += [trans.estim[: self.n_best]]
                bucket_estim += trans.estim[0]
                bucket_words += len(trans.pred_sents[0])
                if "tgt" in batch.keys():
                    bucket_gold_score += trans.gold_score
                    bucket_gold_words += len(trans.gold_sent) + 1

                if self.id_tokenization:
                    n_best_preds = trans.pred_sents[: self.n_best]
                else:
                    n_best_preds = [" ".join(pred) for pred in trans.pred_sents[: self.n_best]]

                if self.report_align:
                    align_pharaohs = [build_align_pharaoh(align) for align in trans.word_aligns[: self.n_best]]
                    n_best_preds_align = [" ".join(align[0]) for align in align_pharaohs]
                    n_best_preds = [
                        pred + DefaultTokens.ALIGNMENT_SEPARATOR + align
                        for pred, align in zip(n_best_preds, n_best_preds_align)
                    ]

                if transform_pipe is not None:
                    n_best_preds = transform_pipe.batch_apply_reverse(n_best_preds)

                bucket_preds += [n_best_preds]

                if self.verbose:
                    srcs = [voc_src[tok] for tok in trans.src[: trans.srclen]]
                    sent_number = next(counter)
                    output = trans.log(sent_number, src_raw=srcs)
                    self._log(output)

                if attn_debug:
                    preds = trans.pred_sents[0]
                    preds.append(DefaultTokens.EOS)
                    attns = trans.attns[0].tolist()
                    if self.data_type == "text":
                        srcs = [voc_src[tok] for tok in trans.src[: trans.srclen].tolist()]
                    else:
                        srcs = [str(item) for item in range(len(attns[0]))]
                    output = report_matrix(srcs, preds, attns)
                    self._log(output)

                if align_debug:
                    if self.gold_align:
                        tgts = trans.gold_sent
                    else:
                        tgts = trans.pred_sents[0]
                    align = trans.word_aligns[0].tolist()
                    if self.data_type == "text":
                        srcs = [voc_src[tok] for tok in trans.src[: trans.srclen].tolist()]
                    else:
                        srcs = [str(item) for item in range(len(align[0]))]
                    output = report_matrix(srcs, tgts, align)
                    self._log(output)

            return (
                bucket_scores,
                bucket_estims,
                bucket_preds,
                bucket_score,
                bucket_estim,
                bucket_words,
                bucket_gold_score,
                bucket_gold_words,
            )

        bucket_predictions = []
        prev_idx = 0

        src_tokens = 0

        for batch, bucket_idx in infer_iter:

            src_tokens += batch["srclen"].sum()
            batch_data = self.predict_batch(batch, attn_debug)

            predictions = prediction_builder.from_batch(batch_data)
            is_seq2seq = (
                getattr(self.model, "encoder", None) is not None and getattr(self.model, "decoder", None) is not None
            )
            if (
                is_seq2seq
                and self._tgt_sep_idx != self._tgt_unk_idx
                and (batch["src"] == self._tgt_sep_idx).any().item()
            ):
                # For seq2seq when we need to force doc to spit the same number of sents
                # TODO: this should be moved to Translator class instead of testing is_seq2seq
                predictions = _maybe_retranslate(predictions, batch)

            bucket_predictions += predictions

            if not isinstance(infer_iter, list) and len(bucket_predictions) >= infer_iter.bucket_size:
                bucket_idx += 1

            if bucket_idx != prev_idx:
                prev_idx = bucket_idx
                (
                    bucket_scores,
                    bucket_estims,
                    bucket_preds,
                    bucket_score,
                    bucket_estim,
                    bucket_words,
                    bucket_gold_score,
                    bucket_gold_words,
                ) = _process_bucket(bucket_predictions)
                all_scores += bucket_scores
                all_estim += bucket_estims
                all_predictions += bucket_preds
                pred_score_total += bucket_score
                estim_total += bucket_estim
                pred_words_total += bucket_words
                gold_score_total += bucket_gold_score
                gold_words_total += bucket_gold_words
                bucket_predictions = []

        if len(bucket_predictions) > 0:
            (
                bucket_scores,
                bucket_estims,
                bucket_preds,
                bucket_score,
                bucket_estim,
                bucket_words,
                bucket_gold_score,
                bucket_gold_words,
            ) = _process_bucket(bucket_predictions)
            all_scores += bucket_scores
            all_estim += bucket_estims
            all_predictions += bucket_preds
            pred_score_total += bucket_score
            estim_total += bucket_estim
            pred_words_total += bucket_words
            gold_score_total += bucket_gold_score
            gold_words_total += bucket_gold_words

        end_time = time()

        if self.report_score:
            msg = self._report_score("PRED", pred_score_total, len(all_scores))
            self._log(msg)
            msg = self._report_score("ESTIM", estim_total, len(all_estim))
            self._log(msg)
            if "tgt" in batch.keys() and not self.tgt_file_prefix:
                msg = self._report_score("GOLD", gold_score_total, len(all_scores))
                self._log(msg)

        if self.report_time:
            total_time = end_time - start_time
            step0 = sum(self.step0_time)
            decoding_time = total_time - step0
            self._log("Step 0 time (s): %.2f" % step0)
            self._log("Enc/Step 0 tokens / sec: %.1f" % (src_tokens / step0))
            self._log("Subsequent prediction time including all (s): %.2f" % decoding_time)
            self._log("Average prediction time (ms): %.1f" % (decoding_time / len(all_predictions) * 1000))
            self._log("Tokens per second: %.1f" % (pred_words_total / decoding_time))
            self._log("pred_words_total: %.1f" % (pred_words_total))

        if self.dump_beam:
            import json

            json.dump(
                self.predictor.beam_accum,
                codecs.open(self.dump_beam, "w", "utf-8"),
            )

        return all_scores, all_estim, all_predictions

    def _score(self, infer_iter):
        self.with_scores = True
        score_res = []
        processed_bucket = {}
        prev_bucket_idx = 0
        for batch, bucket_idx in infer_iter:
            if bucket_idx != prev_bucket_idx:
                prev_bucket_idx += 1
                score_res += [item for _, item in sorted(processed_bucket.items())]
                processed_bucket = {}
            batch_data = self.predict_batch(batch, attn_debug=False, scoring=True)
            batch_gold_scores = batch_data["gold_score"].cpu().numpy().tolist()
            batch_tgt_lengths = batch["tgtlen"].cpu().numpy().tolist()
            batch_inds_in_bucket = batch["ind_in_bucket"]
            if self.return_gold_log_probs:
                batch_gold_log_probs = batch_data["gold_log_probs"].cpu().numpy().tolist()
            else:
                batch_gold_log_probs = [None for i, _ in enumerate(batch_inds_in_bucket)]
            for i, ind in enumerate(batch_inds_in_bucket):
                processed_bucket[ind] = [
                    batch_gold_scores[i],
                    batch_gold_log_probs[i],
                    batch_tgt_lengths[i],
                ]
        if processed_bucket:
            score_res += [item for _, item in sorted(processed_bucket.items())]
        return score_res

    def _align_pad_prediction(self, predictions, bos, pad):
        """
        Padding predictions in batch and add BOS.

        Args:
            predictions (List[List[Tensor]]): `(batch, n_best,)`, for each src
                sequence contain n_best tgt predictions all of which ended with
                eos id.
            bos (int): bos index to be used.
            pad (int): pad index to be used.

        Return:
            batched_nbest_predict (torch.LongTensor): `(batch, n_best, tgt_l)`
        """
        dtype, device = predictions[0][0].dtype, predictions[0][0].device
        flatten_tgt = [best.tolist() for bests in predictions for best in bests]
        paded_tgt = torch.tensor(
            list(zip_longest(*flatten_tgt, fillvalue=pad)),
            dtype=dtype,
            device=device,
        ).T
        bos_tensor = torch.full([paded_tgt.size(0), 1], bos, dtype=dtype, device=device)
        full_tgt = torch.cat((bos_tensor, paded_tgt), dim=-1)
        batched_nbest_predict = full_tgt.view(len(predictions), -1, full_tgt.size(-1))  # (batch, n_best, tgt_l)
        return batched_nbest_predict

    def _report_score(self, name, score_total, nb_sentences):
        # In the case of length_penalty = none we report the total logprobs
        # divided by the number of sentence to get an approximation of the
        # per sentence logprob. We also return the corresponding ppl
        # When a length_penalty is used eg: "avg" or "wu" since logprobs
        # are normalized per token we report the per line per token logprob
        # and the corresponding "per word perplexity"
        if nb_sentences == 0:
            msg = "%s No predictions" % (name,)
        else:
            score = score_total / nb_sentences
            try:
                ppl = exp(-score_total / nb_sentences)
            except OverflowError:
                ppl = float("inf")
            msg = "%s SCORE: %.4f, %s PPL: %.2f NB SENTENCES: %d" % (
                name,
                score,
                name,
                ppl,
                nb_sentences,
            )
        return msg

    def _decode_and_generate(
        self,
        decoder_in,
        enc_out,
        src_len,
        step=None,
        return_attn=False,
        left_pad=False,
        images=None,
    ):

        # Decoder forward, takes [batch, tgt_len, nfeats] as input
        # and [batch, src_len, hidden] as enc_out
        # in case of inference tgt_len = 1, batch = beam times batch_size
        # in case of Gold Scoring tgt_len = actual length, batch = 1 batch
        # we still rely on src_len here because updated at each beam search step
        if isinstance(enc_out, tuple):
            src_max_len = enc_out[0].size(1)
            src_pad_mask = sequence_mask(src_len, src_max_len).unsqueeze(1)  # [B, 1, T_src]
        elif enc_out is not None:
            src_max_len = enc_out.size(1)  # src_len.max() ce bug était sournois
            src_pad_mask = sequence_mask(src_len, src_max_len).unsqueeze(1)  # [B, 1, T_src]
        else:
            src_pad_mask = None

        if images is not None and step == 0:
            emb = self.model.embed_vision_language_features(decoder_in, images=images)
        else:
            emb = self.model.tgt_emb(decoder_in, step=step)

        tgt_pad_mask = decoder_in.eq(self._tgt_pad_idx).unsqueeze(1)  # [B, 1, T_tgt]
        dec_out, dec_attn = self.model.decoder(
            emb,
            enc_out=enc_out,
            src_len=src_len,
            step=step,
            return_attn=self.global_scorer.has_cov_pen or return_attn,
            src_pad_mask=src_pad_mask,
            tgt_pad_mask=tgt_pad_mask,
            left_pad=left_pad,
            decoder_in=decoder_in,
            image_token_id=self.image_token_id,
        )

        # Generator forward.
        if "std" in dec_attn:
            attn = dec_attn["std"]
        else:
            attn = None
        scores = self.model.generator(dec_out.squeeze(1))
        log_probs = log_softmax(scores, dim=-1)  # we keep float16 if FP16
        # returns [(batch_size x beam_size) , vocab ] when 1 step
        # or [batch_size, tgt_len, vocab ] when full sentence
        return log_probs, attn

    def predict_batch(self, batch, attn_debug):
        """Predict a batch of sentences."""
        raise NotImplementedError

    def _score_target(self, batch, enc_out, src_len):
        raise NotImplementedError

    def report_results(
        self,
        gold_score,
        gold_log_probs,
        batch,
        batch_size,
        decode_strategy,
        estim,
    ):
        results = {
            "predictions": None,
            "scores": None,
            "attention": None,
            "estim": None,
            "batch": batch,
            "gold_score": gold_score,
            "gold_log_probs": gold_log_probs,
        }

        if self.report_align:
            results["alignment"] = self._align_forward(batch, decode_strategy.predictions)
        else:
            results["alignment"] = [[] for _ in range(batch_size)]
        preds = []
        scores = []
        attns = []
        sorted_estim = []

        for sublist_estim, sublist_scores, sublist_preds, sublist_attns in zip(
            estim,
            decode_strategy.scores,
            decode_strategy.predictions,
            decode_strategy.attention,
        ):

            # key_fn = lambda x: x[3]  # tri sur estimation
            # key_fn = lambda x: x[0]  # tri sur xentropy

            # use a pareto like instead of one of the score above.
            def key_fn(t):
                CE = float(t[0])
                E = float(t[3])
                # Count how many sentences this sentence dominates within the set
                dominates_count = sum(
                    1
                    for other in zip(sublist_scores, sublist_preds, sublist_attns, sublist_estim)
                    if other is not t
                    and float(other[0]) <= CE
                    and float(other[3]) <= E
                    and (float(other[0]) < CE or float(other[3]) < E)
                )
                # Return tuple: first dominates_count, then E
                return (dominates_count, E)

            sorted_tuples = sorted(
                zip(sublist_scores, sublist_preds, sublist_attns, sublist_estim),
                key=key_fn,
                reverse=True,
            )

            if len(sorted_tuples) == 0:
                scores = decode_strategy.scores
                preds = decode_strategy.predictions
                attns = decode_strategy.attention
                sorted_estim = estim
            else:
                scores.append([item[0] for item in sorted_tuples])
                preds.append([item[1] for item in sorted_tuples])
                attns.append([item[2] for item in sorted_tuples])
                sorted_estim.append([item[3] for item in sorted_tuples])

        results["scores"] = scores
        results["predictions"] = preds
        results["attention"] = attns
        results["estim"] = sorted_estim
        return results



import os
import json
from typing import Dict, List, Any

from eole.config import recursive_model_fields_set
from eole.config.config import get_config_dict
from eole.config.training import TrainingConfig
from eole.config.inference import InferenceConfig
from eole.config.common import MiscConfig, LoggingConfig
from eole.config.models import ModelConfig, build_model_config
from eole.config.data import (
    DataConfig,
    BaseVocabConfig,
    VocabConfig,
    NestedAllTransformsConfig,
)
from eole.transforms import get_transforms_cls
from eole.constants import TransformType
from eole.utils.logging import logger
from pydantic import Field, field_validator, model_validator
import torch


# Models below somewhat replicate prior opt behavior to facilitate transition
class TrainConfig(LoggingConfig, MiscConfig, DataConfig, VocabConfig):  # ModelConfig, TrainingConfig
    n_sample: int = Field(
        default=0,
        description="Number of transformed samples per corpus to use to build the vocabulary. "
        "Set to -1 to use the full corpora.",
    )  # not sure how to handle the legacy build_vocab_only flag here (different default value in both cases) # noqa: E501
    verbose: bool = Field(
        default=False,
        description="Print data loading and statistics for all process " "(default only logs the first process shard).",
    )  # not sure this still works
    model: ModelConfig | None = None  # TypeAdapter handling discrimination directly
    training: TrainingConfig | None = Field(default_factory=TrainingConfig)
    inference: InferenceConfig | None = Field(default=None)

    def get_model_path(self):
        return self.training.get_model_path()

    @property
    def data_type(self):
        return self.training.data_type

    @classmethod
    def get_defaults(cls, architecture):
        return cls(
            src_vocab="dummy",
            tgt_vocab="dummy",
            data={},
            model={"architecture": architecture},
        ).model_dump()

    @field_validator("model", "training", mode="before")
    @classmethod
    def str_to_dict(cls, v) -> Dict:
        if isinstance(v, str):
            if v == "None":
                return None  # not sure about this patch to allow forcing encoder/decoder to None (e.g. LM) # noqa: E501
            w = eval(v)
            return w
        return v

    @model_validator(mode="before")
    @classmethod
    def default_architecture(cls, data: Any) -> Any:
        # only enforce default "custom" if some model settings are passed
        if "model" in data.keys():
            data["model"] = cls.str_to_dict(data["model"])
            if isinstance(data.get("model", {}), dict):
                if data.get("model", {}).get("architecture", None) is None:
                    data["model"]["architecture"] = "custom"
        return data

    @model_validator(mode="after")
    def _validate_train_config(self):
        if self.model is None and self.training.train_from is None:
            raise ValueError(
                "No model architecture is configured. "
                "You should either finetune from an existing model, "
                "or specify a model configuration."
            )
        if self.n_sample != 0:
            assert (
                self.save_data
            ), "-save_data should be set if \
                want save samples."
        # if torch.cuda.is_available() and len(self.training.gpu_ranks) == 0:
        #     logger.warn("You have a CUDA device, should run with -gpu_ranks")
        return self


class PredictConfig(
    InferenceConfig,
    LoggingConfig,
    MiscConfig,
):
    transforms: List[str] | None = []
    transforms_configs: NestedAllTransformsConfig | None = Field(default_factory=NestedAllTransformsConfig)
    # patch to make it work
    share_vocab: bool | None = False
    _all_transform: List[str] | None = None  # might be a cleaner way to do this
    src_subword_vocab: str | None = None  # patch for CT2 inference engine (to improve later)
    model: ModelConfig | None = None
    model_path: str | List[str] = Field(
        description="Path to model .pt file(s). " "Multiple models can be specified for ensemble decoding."
    )  # some specific (mapping to "models") in legacy code, need to investigate
    src: str = Field(description="Source file to decode (one line per sequence).")
    tgt: str | None = Field(
        default=None,
        description="True target sequences, useful for scoring or prefix decoding.",
    )
    tgt_file_prefix: bool = Field(default=False, description="Generate predictions using provided tgt as prefix.")
    output: str = Field(
        default="pred.txt",
        description="Path to output the predictions (each line will be the decoded sequence).",
    )
    engine: str = Field(
        default="eole",
        description="engine to run inference: eole or ct2",
    )

    @model_validator(mode="after")
    def _validate_predict_config(self):
        # Not sure we want to call this at every validation
        self._update_with_model_config()
        self.check_self_attn_backend()
        # TODO: do we really need this _all_transform?
        if self._all_transform is None:
            self._all_transform = self.transforms
        if torch.cuda.is_available() and len(self.gpu_ranks) == 0:
            logger.warn("You have a CUDA device, should run with -gpu_ranks")
        return self

    def _update_with_model_config(self):
        # Note: in case of ensemble decoding, grabbing the first model's
        # config and artifacts by default
        os.environ["MODEL_PATH"] = self.model_path[0]
        config_path = os.path.join(self.model_path[0], "config.json")
        if os.path.exists(config_path):
            with open(config_path) as f:
                config_dict = json.loads(os.path.expandvars(f.read()))
        else:
            config_dict = {}
        # Filter out Train transforms
        transforms = config_dict.get("transforms", [])
        transforms_cls = get_transforms_cls(transforms)
        transforms = [t for t in transforms if transforms_cls[t].type != TransformType.Train]

        if os.path.exists(config_path):
            # logic from models.BaseModel.inference_logic
            model_config = build_model_config(config_dict.get("model", {}))
            training_config = TrainingConfig(**config_dict.get("training", {}))
            training_config.world_size = self.world_size
            training_config.gpu_ranks = self.gpu_ranks
            # retrieve share_vocab from checkpoint config
            self.__dict__["share_vocab"] = config_dict.get("share_vocab", False)
            # retrieve precision from checkpoint config if not explicitly set
            if "compute_dtype" not in self.model_fields_set:
                self.compute_dtype = training_config.compute_dtype
            # quant logic, might be better elsewhere
            if hasattr(training_config, "quant_type") and training_config.quant_type in [
                "awq_gemm",
                "awq_gemv",
            ]:
                if (
                    hasattr(self, "quant_type")
                    and self.quant_type != ""
                    and self.quant_type != training_config.quant_type
                ):
                    raise ValueError("Model is a awq quantized model, cannot overwrite with another quant method")
                self.update(quant_type=training_config.quant_type)
            elif self.quant_type == "" and training_config.quant_type != "":
                self.update(
                    quant_layers=training_config.quant_layers,
                    quant_type=training_config.quant_type,
                )

            self.update(
                model=model_config,
            )

        update_dict = {}
        if "transforms" not in self.model_fields_set:
            update_dict["transforms"] = transforms
            update_dict["_all_transform"] = transforms
        if "transforms_configs" not in self.model_fields_set:
            update_dict["transforms_configs"] = NestedAllTransformsConfig(**config_dict.get("transforms_configs", {}))
        else:
            # merge model transforms_configs with explicitly set config
            # useful for cases like prefix/suffix that might not be set in the checkpoint
            transforms_configs = recursive_model_fields_set(self.transforms_configs)
            transforms_configs.update(config_dict.get("transforms_configs", {}))
            update_dict["transforms_configs"] = NestedAllTransformsConfig(**transforms_configs)
        if "compute_dtype" not in self.model_fields_set:
            self.compute_dtype = config_dict.get("training", {}).get("compute_dtype", "fp16")
        for key, value in config_dict.get("inference", {}).items():
            if key not in self.model_fields_set:
                update_dict[key] = value
        self.update(**update_dict)


PredictConfig.model_rebuild()


class BuildVocabConfig(DataConfig, MiscConfig, BaseVocabConfig):  # , AllTransformsConfig):

    model_config = get_config_dict()
    model_config["extra"] = "ignore"

    # some specific items related to build_vocab_only flag
    dump_samples: bool = Field(
        default=False,
        description="Dump samples when building vocabulary. " "Warning: this may slow down the process.",
    )
    num_threads: int = Field(default=1, description="Number of parallel threads to build the vocabulary.")
    learn_subwords: bool = Field(
        default=False,
        description="Learn subwords (based on defined transforms) prior to building vocabulary.",
    )
    learn_subwords_size: int = Field(default=32000, description="Number of subwords operations to learn.")
    vocab_sample_queue_size: int = Field(default=20, description="Size of queues used for dumping samples.")
    n_sample: int = Field(
        default=5000,
        description="Number of transformed samples per corpus to use to build the vocabulary. "
        "Set to -1 to use the full corpora.",
    )  # not sure how to handle the legacy build_vocab_only flag here (different default value in both cases) # noqa: E501

    # We might want to validate that we have onmt_tokenize transform if learn_subwords...
    @model_validator(mode="after")
    def _validate_build_vocab_config(self):
        if self.learn_subwords:
            assert "onmt_tokenize" in self.transforms, (
                "The onmt_tokenizer transform should be set " "and configured to enable learn_subwords."
            )



import os
import torch
from collections import OrderedDict
from eole.config.config import Config
from eole.utils.logging import logger


# default EOLE_MODEL_DIR
if os.environ.get("EOLE_MODEL_DIR", None) is None:
    os.environ["EOLE_MODEL_DIR"] = os.getcwd()


def calculate_depth(value, current_depth=0):
    if isinstance(value, (dict, Config)):
        if isinstance(value, Config):
            value = value.__dict__
        return max(calculate_depth(v, current_depth + 1) for v in value.values()) if value else current_depth
    return current_depth


def reorder_fields(fields):
    """
    Put non nested fields before nested ones in config json dump,
    for better readability.
    """
    ordered_fields = OrderedDict()
    non_nested_fields = []
    nested_fields = []

    for field, value in fields.items():
        if isinstance(value, dict):
            nested_fields.append((field, value))
        else:
            non_nested_fields.append((field, value))

    # Add non-nested fields first
    for field, value in non_nested_fields:
        ordered_fields[field] = value

    # Calculate depths and sort nested fields
    nested_fields.sort(key=lambda x: calculate_depth(x[1]))

    # Add nested fields
    for field, value in nested_fields:
        ordered_fields[field] = reorder_fields(value)

    return ordered_fields


def recursive_model_fields_set(model):
    fields = {}
    if isinstance(model, dict):
        fields_to_check = model.keys()
    else:
        # we need to explicitly handle "architecture" related fields for now
        # to properly handle discriminated config instanciation
        discriminator_fields = {
            "architecture",
            "encoder_type",
            "decoder_type",
        }.intersection(set(model.__fields__.keys()))
        nested_models = {key for key, value in model.__dict__.items() if isinstance(value, Config)}
        fields_to_check = model.model_fields_set | discriminator_fields | nested_models
    for field in fields_to_check:
        if isinstance(model, dict):
            field_value = model.get(field, None)
        else:
            field_value = getattr(model, field, None)
        if isinstance(field_value, Config) or isinstance(field_value, dict):
            _fields = recursive_model_fields_set(field_value)
            if _fields != {}:
                fields[field] = _fields
        else:
            if isinstance(field_value, torch.dtype):
                # torch.dtype is not serializable
                field_value = str(field_value)
            fields[field] = field_value
    return reorder_fields(fields)


def recursive_update_dict(_dict, new_dict, defaults):
    # patch to allow populating previously none keys
    # (e.g. data in finetuned HF converted model)
    if _dict is None:
        _dict = {}

    for k, v in new_dict.items():
        if k == "architecture":
            # keep info from checkpoint
            continue
        if isinstance(v, dict):
            _dict[k] = recursive_update_dict(_dict.get(k, {}), v, defaults.get(k, {}))
        else:
            default = defaults.get(k, None) if isinstance(defaults, dict) else None
            previous_v = _dict.get(k, default)
            if v != previous_v:
                logger.info(f"Option: {k}, value: {v}, overriding model: {previous_v}")
            _dict[k] = v
    return _dict


def get_non_default_values(parsed_args, defaults):
    non_default_values = {}
    # defaults = vars(parse_args())
    for key, value in vars(parsed_args).items():
        if value != defaults.get(key, None):
            non_default_values[key] = value
    return non_default_values



""" Modules for prediction """

from eole.predict.translator import Translator
from eole.predict.generator import GeneratorLM
from eole.predict.encoder import Encoder

from eole.predict.beam_search import GNMTGlobalScorer
from eole.decoders.ensemble import load_test_model as ensemble_load_test_model
from eole.models.model import BaseModel


def get_infer_class(model_config):
    # might have more cases later
    if model_config.decoder is None:
        return Encoder
    elif model_config.encoder is None:
        return GeneratorLM
    elif model_config.encoder.encoder_type == "vision":
        return GeneratorLM
    else:
        return Translator


def build_predictor(config, device_id=0, report_score=True, logger=None):
    # right now config is a full (non nested) PredictConfig

    load_test_model = ensemble_load_test_model if len(config.model_path) > 1 else BaseModel.load_test_model

    vocabs, model, model_config = load_test_model(config, device_id)
    config.update(model=model_config)

    scorer = GNMTGlobalScorer.from_config(config)

    infer_class = get_infer_class(model_config)
    if infer_class is not None:
        translator = infer_class.from_config(
            model,
            vocabs,
            config,
            model_config,
            device_id=device_id,
            global_scorer=scorer,
            report_align=config.report_align,
            report_score=report_score,
            logger=logger,
        )
    else:
        raise NotImplementedError(f"No infer_class found for {model_config.__class__}")
    return translator



import unittest
from eole.modules.embeddings import Embeddings
from eole.constants import PositionEncodingType

import itertools
from copy import deepcopy

import torch
import random
from torch.nn.init import xavier_uniform_
from eole.tests.utils_for_tests import product_dict


class TestEmbeddings(unittest.TestCase):
    INIT_CASES = list(
        product_dict(
            word_vec_size=[172],
            word_vocab_size=[319],
            word_padding_idx=[17],
            position_encoding_type=[PositionEncodingType.SinusoidalInterleaved, None],
            dropout=[0, 0.5],
            freeze_word_vecs=[False, True],
        )
    )
    PARAMS = list(product_dict(batch_size=[1, 14], max_seq_len=[23]))

    @classmethod
    def case_is_degenerate(cls, case):
        return False

    @classmethod
    def cases(cls):
        for case in cls.INIT_CASES:
            if not cls.case_is_degenerate(case):
                yield case

    @classmethod
    def dummy_inputs(cls, params, init_case):
        max_seq_len = params["max_seq_len"]
        batch_size = params["batch_size"]
        n_words = init_case["word_vocab_size"]
        voc_sizes = [n_words]
        pad_idxs = [init_case["word_padding_idx"]]
        lengths = [random.randint(0, max_seq_len) for _ in range(batch_size)]
        lengths[0] = max_seq_len
        inps = torch.empty((batch_size, max_seq_len), dtype=torch.long)
        for voc_size, pad_idx in zip(voc_sizes, pad_idxs):
            for b, len_ in enumerate(lengths):
                inps[b, :len_] = torch.randint(0, voc_size - 1, (len_,))
                inps[b, len_:] = pad_idx
        return inps

    @classmethod
    def expected_shape(cls, params, init_case):
        wvs = init_case["word_vec_size"]
        size = wvs
        return params["batch_size"], params["max_seq_len"], size

    def test_embeddings_forward_shape(self):
        for params, init_case in itertools.product(self.PARAMS, self.cases()):
            emb = Embeddings(**init_case)
            dummy_in = self.dummy_inputs(params, init_case)
            res = emb(dummy_in)
            expected_shape = self.expected_shape(params, init_case)
            self.assertEqual(res.shape, expected_shape, init_case.__str__())

    def test_embeddings_trainable_params(self):
        for params, init_case in itertools.product(self.PARAMS, self.cases()):
            emb = Embeddings(**init_case)
            trainable_params = {n: p for n, p in emb.named_parameters() if p.requires_grad}
            # first check there's nothing unexpectedly not trainable
            for key in emb.state_dict():
                if key not in trainable_params:
                    if key.endswith("embeddings.weight") and init_case["freeze_word_vecs"]:
                        # ok: word embeddings shouldn't be trainable
                        # if word vecs are freezed
                        continue
                    if key.endswith("pe.pe"):
                        # ok: positional encodings shouldn't be trainable
                        assert init_case["position_encoding_type"]
                        continue
                    else:
                        self.fail("Param {:s} is unexpectedly not " "trainable.".format(key))
            # then check nothing unexpectedly trainable
            if init_case["freeze_word_vecs"]:
                self.assertFalse(
                    any(trainable_param.endswith("embeddings.weight") for trainable_param in trainable_params),
                    "Word embedding is trainable but word vecs are freezed.",
                )
            if init_case["position_encoding_type"]:
                self.assertFalse(
                    any(trainable_p.endswith(".pe.pe") for trainable_p in trainable_params),
                    "Positional encoding is trainable.",
                )

    def test_embeddings_trainable_params_update(self):
        for params, init_case in itertools.product(self.PARAMS, self.cases()):
            emb = Embeddings(**init_case)
            for n, p in emb.named_parameters():
                if "weight" in n and p.dim() > 1:
                    xavier_uniform_(p)
            trainable_params = {n: p for n, p in emb.named_parameters() if p.requires_grad}
            if len(trainable_params) > 0:
                old_weights = deepcopy(trainable_params)
                dummy_in = self.dummy_inputs(params, init_case)
                res = emb(dummy_in)
                pretend_loss = res.mean()
                pretend_loss.backward()
                dummy_optim = torch.optim.SGD(trainable_params.values(), 1)
                dummy_optim.step()
                for param_name in old_weights.keys():
                    self.assertTrue(
                        trainable_params[param_name].ne(old_weights[param_name]).any(),
                        param_name + " " + init_case.__str__(),
                    )



import torch
from eole.predict.inference import Inference
from eole.constants import ModelType
from eole.predict.greedy_search import GreedySearch
from eole.predict.beam_search import BeamSearch


class Encoder(Inference):
    @classmethod
    def validate_task(cls, task):
        if task != ModelType.ENCODER:
            raise ValueError(f"Encoder does not support task {task}." f" Tasks supported: {ModelType.ENCODER}")

    def predict_batch(self, batch, attn_debug):
        """Predict a batch of sentences."""
        if self.max_length_ratio > 0:
            max_length = int(min(self.max_length, batch["src"].size(1) * self.max_length_ratio + 5))
        else:
            max_length = self.max_length
        with torch.no_grad():
            if self.top_k != 0 or self.top_p != 0:
                decode_strategy = GreedySearch(
                    pad=self._tgt_pad_idx,
                    bos=self._tgt_bos_idx,
                    eos=self._tgt_eos_idx,
                    unk=self._tgt_unk_idx,
                    start=self._tgt_start_with,
                    n_best=self.n_best,
                    batch_size=len(batch["srclen"]),
                    global_scorer=self.global_scorer,
                    min_length=self.min_length,
                    max_length=max_length,
                    block_ngram_repeat=self.block_ngram_repeat,
                    exclusion_tokens=self._exclusion_idxs,
                    return_attention=attn_debug or self.replace_unk,
                    sampling_temp=self.temperature,
                    top_k=self.top_k,
                    top_p=self.top_p,
                    beam_size=self.beam_size,
                    ban_unk_token=self.ban_unk_token,
                    add_estimator=self.add_estimator,
                )
            else:
                # TODO: support these blacklisted features
                assert not self.dump_beam
                decode_strategy = BeamSearch(
                    self.beam_size,
                    batch_size=len(batch["srclen"]),
                    pad=self._tgt_pad_idx,
                    bos=self._tgt_bos_idx,
                    eos=self._tgt_eos_idx,
                    unk=self._tgt_unk_idx,
                    start=self._tgt_start_with,
                    n_best=self.n_best,
                    global_scorer=self.global_scorer,
                    min_length=self.min_length,
                    max_length=max_length,
                    return_attention=attn_debug or self.replace_unk,
                    block_ngram_repeat=self.block_ngram_repeat,
                    exclusion_tokens=self._exclusion_idxs,
                    stepwise_penalty=self.stepwise_penalty,
                    ratio=self.ratio,
                    ban_unk_token=self.ban_unk_token,
                    add_estimator=self.add_estimator,
                )
            return self._predict_batch_with_strategy(batch, decode_strategy)

    def _run_encoder(self, batch):
        src = batch["src"]
        src_len = batch["srclen"]
        batch_size = len(batch["srclen"])
        emb = self.model.src_emb(src)
        pad_mask = src.eq(self._src_pad_idx).unsqueeze(1)
        enc_out, enc_final_hs = self.model.encoder(emb, pad_mask=pad_mask)

        if src_len is None:
            assert not isinstance(enc_out, tuple), "Ensemble decoding only supported for text data"
            src_len = torch.Tensor(batch_size).type_as(enc_out).long().fill_(enc_out.size(1))
        return src, enc_final_hs, enc_out, src_len

    def _predict_batch_with_strategy(self, batch, decode_strategy):
        """Predict a batch of sentences step by step using cache.

        Args:
            batch: a batch of sentences, yield by data iterator.
            decode_strategy (DecodeStrategy): A decode strategy to use for
                generate prediction step by step.

        Returns:
            results (dict): The prediction results.
        """
        # (0) Prep the components of the search.
        batch_size = len(batch["srclen"])

        # (1) Run the encoder on the src.
        src, enc_final_hs, enc_out, src_len = self._run_encoder(batch)

        gold_score, gold_log_probs = self._gold_score(
            batch,
            enc_out,
            src_len,
            enc_final_hs,
            batch_size,
            src,
        )

        if self.add_estimator:
            """
            # Version with encoder out average
            pad_mask1 = ~src.eq(self._tgt_pad_idx)
            in_estim1 = (enc_out * pad_mask1.unsqueeze(-1).float()).sum(
                dim=1
            ) / pad_mask1.sum(dim=1, keepdim=True).float()
            estim = self.model.estimator(in_estim1.half()).squeeze(-1)
            """
            # Version with first token embedding (same as COMET)
            estim = self.model.estimator(enc_out[:, 0, :]).squeeze(-1)
        else:
            estim = torch.ones([enc_out.size(0)])
        estim = [[item] for item in estim.tolist()]

        results = {
            "predictions": [[torch.tensor([0])] for _ in range(batch_size)],
            "scores": [[0]] * batch_size,
            "attention": [[0]] * batch_size,
            "estim": estim,
            "batch": batch,
            "gold_score": [0] * batch_size,
            "gold_log_probs": None,
            "alignment": [[] for _ in range(batch_size)],
        }

        return results

    def _score_target(self, batch, enc_out, src_len):
        tgt = batch["tgt"]
        tgt_in = tgt[:, :-1, :]

        log_probs, attn = self._decode_and_generate(
            tgt_in,
            enc_out,
            src_len=src_len,
        )

        log_probs[:, :, self._tgt_pad_idx] = 0
        gold = tgt[:, 1:, :]
        gold_scores = log_probs.gather(2, gold)
        gold_scores = gold_scores.sum(dim=1).view(-1)
        return gold_scores, None



import torch
import torch.nn as nn

from eole.decoders.decoder import DecoderBase
from eole.modules.stacked_rnn import StackedLSTM, StackedGRU
from eole.modules.gate import context_gate_factory
from eole.modules.global_attention import GlobalAttention


class RNNDecoderBase(DecoderBase):
    """Base recurrent attention-based decoder class.

    Specifies the interface used by different decoder types
    and required by :class:`~eole.models.BaseModel`.

    Args:
        decoder_config (eole.config.DecoderConfig): full decoder config
        running_config (TrainingConfig / InferenceConfig)
    """

    def __init__(
        self,
        decoder_config,
        running_config=None,
        with_cross_attn=False,
    ):
        super(RNNDecoderBase, self).__init__(
            attentional=decoder_config.global_attention != "none" and decoder_config.global_attention is not None
        )

        self.bidirectional_encoder = decoder_config.bidirectional_encoder
        self.num_layers = decoder_config.layers
        self.dropout = nn.Dropout(getattr(running_config, "dropout", [0.0])[0])

        # Build the RNN.
        self.rnn = self._build_rnn(
            decoder_config.rnn_type,
            input_size=self._input_size,
            hidden_size=self.hidden_size,
            num_layers=self.num_layers,
            dropout=getattr(running_config, "dropout", [0.0])[0],
        )

        # Set up the context gate.
        self.context_gate = None
        if decoder_config.context_gate is not None:
            self.context_gate = context_gate_factory(
                decoder_config.context_gate,
                self._input_size,
                self.hidden_size,
                self.hidden_size,
                self.hidden_size,
            )

        # Set up the standard attention.
        self._coverage = decoder_config.coverage_attn
        if not self.attentional:
            if self._coverage:
                raise ValueError("Cannot use coverage term with no attention.")
            self.attn = None
        else:
            self.attn = GlobalAttention(
                self.hidden_size,
                coverage=decoder_config.coverage_attn,
                attn_type=decoder_config.global_attention,
                attn_func=decoder_config.global_attention_function,
            )

    @classmethod
    def from_config(cls, decoder_config, running_config=None, with_cross_attn=False):
        """Alternate constructor."""
        # config = opt.model.decoder  # RnnDecoderConfig
        return cls(
            decoder_config,
            running_config=running_config,
            with_cross_attn=False,
        )

    def init_state(self, **kwargs):
        """Initialize decoder state with last state of the encoder."""
        enc_final_hs = kwargs.pop("enc_final_hs", None)

        def _fix_enc_hidden(hidden):
            # The encoder hidden is  (layers*directions) x batch x dim.
            # We need to convert it to layers x batch x (directions*dim).
            if self.bidirectional_encoder:
                hidden = torch.cat([hidden[0 : hidden.size(0) : 2], hidden[1 : hidden.size(0) : 2]], 2)
            return hidden

        if isinstance(enc_final_hs, tuple):  # LSTM
            self.state["hidden"] = tuple(_fix_enc_hidden(enc_hid) for enc_hid in enc_final_hs)
        else:  # GRU
            self.state["hidden"] = (_fix_enc_hidden(enc_final_hs),)

        # Init the input feed.
        batch_size = self.state["hidden"][0].size(1)

        h_size = (batch_size, self.hidden_size)
        self.state["input_feed"] = self.state["hidden"][0].data.new(*h_size).zero_().unsqueeze(0)

        self.state["coverage"] = None

    def map_state(self, fn):
        # map_state applies on dim=0 (batch_size)
        # TODO refactor RNN states so that batch_size is natively on dim=0
        self.state["hidden"] = tuple(fn(h.transpose(0, 1)).transpose(0, 1) for h in self.state["hidden"])
        self.state["input_feed"] = fn(self.state["input_feed"].transpose(0, 1)).transpose(0, 1)
        if self._coverage and self.state["coverage"] is not None:
            self.state["coverage"] = fn(self.state["coverage"].transpose(0, 1)).transpose(0, 1)

    def detach_state(self):
        self.state["hidden"] = tuple(h.detach() for h in self.state["hidden"])
        self.state["input_feed"] = self.state["input_feed"].detach()
        if self._coverage and self.state["coverage"] is not None:
            self.state["coverage"] = self.state["coverage"].detach()

    def forward(self, emb, enc_out, src_len=None, step=None, **kwargs):
        """
        Args:
            emb (FloatTensor): input embeddings
                 ``(batch, tgt_len, dim)``.
            enc_out (FloatTensor): vectors from the encoder
                 ``(batch, src_len, hidden)``.
            src_len (LongTensor): the padded source lengths
                ``(batch,)``.

        Returns:
            (FloatTensor, dict[str, FloatTensor]):

            * dec_outs: output from the decoder (after attn)
              ``(batch, tgt_len, hidden)``.
            * attns: distribution over src at each tgt
              ``(batch, tgt_len, src_len)``.
        """
        dec_state, dec_outs, attns = self._run_forward_pass(emb, enc_out, src_len=src_len)

        # Update the state with the result.
        if not isinstance(dec_state, tuple):
            dec_state = (dec_state,)
        self.state["hidden"] = dec_state

        # Concatenates sequence of tensors along a new dimension.
        # NOTE: v0.3 to 0.4: dec_outs / attns[*] may not be list
        #       since stack(Variable) was allowed.
        if isinstance(dec_outs, list):
            dec_outs = torch.stack(dec_outs, dim=1)
            for k in attns:
                if isinstance(attns[k], list):
                    attns[k] = torch.stack(attns[k])

        self.state["input_feed"] = dec_outs[:, -1, :].unsqueeze(0)
        self.state["coverage"] = None
        if "coverage" in attns:
            self.state["coverage"] = attns["coverage"][-1, :, :].unsqueeze(0)

        return dec_outs, attns

    def update_dropout(self, dropout, attention_dropout=None):
        self.dropout.p = dropout


class StdRNNDecoder(RNNDecoderBase):
    """Standard fully batched RNN decoder with attention.

    Faster implementation, uses CuDNN for implementation.
    See :class:`~eole.decoders.decoder.RNNDecoderBase` for options.


    Based around the approach from
    "Neural Machine Translation By Jointly Learning To Align and Translate"
    :cite:`Bahdanau2015`


    Implemented without input_feeding and currently with no `coverage_attn`
    """

    def __init__(
        self,
        decoder_config,
        running_config=None,
        with_cross_attn=False,
    ):
        self.hidden_size = decoder_config.hidden_size
        self._input_size = decoder_config.tgt_word_vec_size
        super(StdRNNDecoder, self).__init__(decoder_config, running_config)

    def _run_forward_pass(self, emb, enc_out, src_len=None):
        """
        Private helper for running the specific RNN forward pass.
        Must be overriden by all subclasses.

        Args:
            emb (FloatTensor): input embeddings
                ``(batch, tgt_len, dim)``.
            enc_out (FloatTensor): output(tensor sequence) from the
                encoder RNN of size ``(batch, src_len, hidden_size)``.
            src_len (LongTensor): the source enc_out lengths.

        Returns:
            (Tensor, List[FloatTensor], Dict[str, List[FloatTensor]):

            * dec_state: final hidden state from the decoder.
            * dec_outs: an array of output of every time
              step from the decoder.
            * attns: a dictionary of different
              type of attention Tensor array of every time
              step from the decoder.
        """

        assert not self._coverage  # TODO, no support yet.

        attns = {}

        if isinstance(self.rnn, nn.GRU):
            rnn_out, dec_state = self.rnn(emb, self.state["hidden"][0])
        else:
            rnn_out, dec_state = self.rnn(emb, self.state["hidden"])

        tgt_batch, tgt_len, _ = emb.size()

        # Calculate the attention.
        if not self.attentional:
            dec_outs = rnn_out
        else:
            dec_outs, p_attn = self.attn(rnn_out, enc_out, src_len=src_len)
            attns["std"] = p_attn

        # Calculate the context gate.
        if self.context_gate is not None:
            dec_outs = self.context_gate(
                emb.view(-1, emb.size(2)),
                rnn_out.view(-1, rnn_out.size(2)),
                dec_outs.view(-1, dec_outs.size(2)),
            )
            dec_outs = dec_outs.view(tgt_batch, tgt_len, self.hidden_size)

        dec_outs = self.dropout(dec_outs)

        return dec_state, dec_outs, attns

    def _build_rnn(self, rnn_type, **kwargs):
        rnn = getattr(nn, rnn_type)(batch_first=True, **kwargs)
        return rnn


class InputFeedRNNDecoder(RNNDecoderBase):
    """Input feeding based decoder.

    See :class:`~eole.decoders.decoder.RNNDecoderBase` for options.

    Based around the input feeding approach from
    "Effective Approaches to Attention-based Neural Machine Translation"
    :cite:`Luong2015`

    """

    def __init__(
        self,
        decoder_config,
        running_config=None,
        with_cross_attn=False,
    ):
        self.hidden_size = decoder_config.hidden_size
        self._input_size = decoder_config.tgt_word_vec_size + self.hidden_size
        super(InputFeedRNNDecoder, self).__init__(decoder_config, running_config)

    def _run_forward_pass(self, emb, enc_out, src_len=None):
        """
        See StdRNNDecoder._run_forward_pass() for description
        of arguments and return values.
        """
        # Additional args check.
        input_feed = self.state["input_feed"].squeeze(0)

        dec_outs = []
        attns = {}
        if self.attn is not None:
            attns["std"] = []
        if self._coverage:
            attns["coverage"] = []

        assert emb.dim() == 3  # batch x len x embedding_dim

        dec_state = self.state["hidden"]

        coverage = self.state["coverage"].squeeze(0) if self.state["coverage"] is not None else None

        # Input feed concatenates hidden state with
        # input at every time step.
        for emb_t in emb.split(1, dim=1):
            dec_in = torch.cat([emb_t.squeeze(1), input_feed], 1)
            rnn_out, dec_state = self.rnn(dec_in, dec_state)
            if self.attentional:
                dec_out, p_attn = self.attn(rnn_out, enc_out, src_len=src_len)
                attns["std"].append(p_attn)
            else:
                dec_out = rnn_out
            if self.context_gate is not None:
                # TODO: context gate should be employed
                # instead of second RNN transform.
                dec_out = self.context_gate(dec_in, rnn_out, dec_out)
            dec_out = self.dropout(dec_out)
            input_feed = dec_out

            dec_outs += [dec_out]

            # Update the coverage attention.
            # attns["coverage"] is actually c^(t+1) of See et al(2017)
            # 1-index shifted
            if self._coverage:
                coverage = p_attn if coverage is None else p_attn + coverage
                attns["coverage"] += [coverage]

        return dec_state, dec_outs, attns

    def _build_rnn(self, rnn_type, input_size, hidden_size, num_layers, dropout):
        stacked_cell = StackedLSTM if rnn_type == "LSTM" else StackedGRU
        return stacked_cell(num_layers, input_size, hidden_size, dropout)

    def update_dropout(self, dropout, attention_dropout=None):
        self.dropout.p = dropout
        self.rnn.dropout.p = dropout



import torch

# import torch.nn.functional as F
# import math
from eole.constants import DefaultTokens, CorpusTask, ModelType
from torch.nn.utils.rnn import pad_sequence
from eole.utils.logging import logger


def text_sort_key(ex):
    """Sort using the number of tokens in the sequence."""
    if ex.get("tgt", None) is not None:
        return len(ex["src"]["src_ids"]), len(ex["tgt"]["tgt_ids"])
    return len(ex["src"]["src_ids"])


def clean_example(maybe_example):
    if isinstance(maybe_example["src"], list):
        maybe_example["src"] = {"src": " ".join(maybe_example["src"])}
    else:
        maybe_example["src"] = {"src": maybe_example["src"]}
    if maybe_example.get("tgt", None) is not None:
        maybe_example["tgt"] = {"tgt": " ".join(maybe_example["tgt"])}
    if "align" in maybe_example:
        maybe_example["align"] = " ".join(maybe_example["align"])
    if "sco" not in maybe_example:
        maybe_example["sco"] = 1
    return maybe_example


def transform_bucket(task, bucket, threshold=0):
    """Returns valid transformed bucket from bucket."""
    transform_cid_to_examples = {}
    for example in bucket:
        transform_cid = (example[1], example[2])
        if transform_cid not in transform_cid_to_examples:
            transform_cid_to_examples[transform_cid] = []
        transform_cid_to_examples[transform_cid].append(example)

    transformed_bucket = []
    # careful below it will return a bucket sorted by corpora
    # but we sort by length later and shuffle batches
    for (transform, cid), sub_bucket in transform_cid_to_examples.items():
        transf_bucket = transform.batch_apply(sub_bucket, is_train=(task == CorpusTask.TRAIN), corpus_name=cid)
        for example, transform, cid in transf_bucket:
            example = clean_example(example)
            if len(example["src"]["src"]) > 0 and example.get("sco", 1) >= threshold:
                transformed_bucket.append(example)

        # at this point an example looks like:
        # {'src': {'src': ..., 'feats': [....]},
        #  'tgt': {'tgt': ...},
        #  'src_original': ['tok1', ...'tokn'],
        #  'tgt_original': ['tok1', ...'tokm'],
        #  'cid': corpus id
        #  'cid_line_number' : cid line number
        #  'align': ...,
        #  if tokenize_id is used then it will include:
        #  'src_ids' and 'tgt_ids'
        # }
    if len(transformed_bucket) > 0:
        return transformed_bucket
    else:
        return None


def numericalize(vocabs, example, model_type=ModelType.ENCODER_DECODER, task=CorpusTask.INFER):
    """ """
    decoder_start_token = vocabs["decoder_start_token"]
    numeric = example

    # In the case transform `tokenizer_id` is used ie hugging_face tokenizer
    # fields `src_ids` and potentially `tgt_ids` are already filled
    numeric["src"]["src_ids"] = example.pop("src_ids", [])
    maybe_tgt_ids = example.pop("tgt_ids", [])

    if model_type == ModelType.ENCODER_DECODER:
        src_text = example["src"]["src"].split(" ")
        if numeric["src"]["src_ids"] == []:
            numeric["src"]["src_ids"] = vocabs["src"](src_text)
        if example.get("tgt", None) is not None:
            if maybe_tgt_ids != []:
                # TODO: handle this better in HF tokenizer templates
                if decoder_start_token != "":
                    decoder_start_token_id = vocabs["tgt"].tokens_to_ids[decoder_start_token]
                    if maybe_tgt_ids[0] != decoder_start_token_id:
                        maybe_tgt_ids = [decoder_start_token_id] + maybe_tgt_ids
                numeric["tgt"]["tgt_ids"] = maybe_tgt_ids
            else:
                tgt_text = example["tgt"]["tgt"].split(" ")
                numeric["tgt"]["tgt_ids"] = vocabs["tgt"](
                    [decoder_start_token] + tgt_text + [vocabs["specials"].get("eos_token", "")]
                )

    elif model_type == ModelType.DECODER:
        if numeric["src"]["src_ids"] == []:
            # tokenize_id not used, we add start_token and numericalize
            src_text = example["src"]["src"].split(" ")
            if decoder_start_token != "":
                src_text = [decoder_start_token] + src_text
            numeric["src"]["src_ids"] = vocabs["src"](src_text)
        numeric["src"]["prefix_len"] = len(numeric["src"]["src_ids"])
        if example["tgt"] is None:
            # no path_tgt given need to use src, add eos if any and remove bos if any
            if task in [CorpusTask.TRAIN, CorpusTask.VALID]:
                example["tgt"] = {}
                if maybe_tgt_ids != []:
                    # tokenize_id in use
                    numeric["tgt"]["tgt_ids"] = maybe_tgt_ids
                else:
                    tgt_text = example["src"]["src"].split(" ")
                    numeric["tgt"]["tgt_ids"] = vocabs["tgt"](tgt_text + [vocabs["specials"].get("eos_token", "")])
                if decoder_start_token == "" or len(numeric["tgt"]["tgt_ids"]) != len(numeric["src"]["src_ids"]):
                    numeric["tgt"]["tgt_ids"] = numeric["tgt"]["tgt_ids"][1:]
        else:
            # path_tgt given, prompt from path_src and answer from path_tgt
            # we concat both, add eos if any on tgt side and remove bos if any.
            if maybe_tgt_ids != []:
                numeric["tgt"]["tgt_ids"] = numeric["src"]["src_ids"] + maybe_tgt_ids
                numeric["src"]["src_ids"] = numeric["src"]["src_ids"] + maybe_tgt_ids[:-1]
            else:
                tgt_text = example["tgt"]["tgt"].split(" ")
                numeric["tgt"]["tgt_ids"] = vocabs["tgt"](tgt_text + [vocabs["specials"].get("eos_token", "")])
                numeric["src"]["src_ids"] = numeric["src"]["src_ids"] + numeric["tgt"]["tgt_ids"]
                numeric["tgt"]["tgt_ids"] = numeric["src"]["src_ids"]
                numeric["src"]["src_ids"] = numeric["src"]["src_ids"][:-1]
            numeric["tgt"]["tgt_ids"] = numeric["tgt"]["tgt_ids"][1:]

    # TODO: support id tokenization
    elif model_type == ModelType.ENCODER:
        src_text = example["src"]["src"].split(" ")
        if example["tgt"] is not None:  # TO BE DISCUSSED
            tgt_text = example["tgt"]["tgt"].split(" ")
            txt = (
                [vocabs["specials"].get("bos_token", "")]
                + tgt_text
                + [vocabs["specials"].get("eos_token", "")]
                + [vocabs["specials"].get("eos_token", "")]
                + src_text
                + [vocabs["specials"].get("eos_token", "")]
            )
            numeric["src"]["src_ids"] = vocabs["src"](txt)
            numeric["tgt"]["tgt_ids"] = vocabs["src"](txt)
        else:
            txt = [vocabs["specials"].get("bos_token", "")] + src_text + [vocabs["specials"].get("eos_token", "")]
            numeric["src"]["src_ids"] = vocabs["src"](txt)

    else:
        raise ValueError(f"Something went wrong with model_type {model_type}")

    return numeric


def parse_align_idx(align_pharaoh):
    """
    Parse Pharaoh alignment into [[<src>, <tgt>], ...]
    """
    align_list = align_pharaoh.strip().split(" ")
    flatten_align_idx = []
    for align in align_list:
        try:
            src_idx, tgt_idx = align.split("-")
        except ValueError:
            logger.warning("{} in `{}`".format(align, align_pharaoh))
            logger.warning("Bad alignement line exists. Please check file!")
            raise
        flatten_align_idx.append([int(src_idx), int(tgt_idx)])
    return flatten_align_idx


def tensorify(vocabs, minibatch, device, left_pad=False):
    """
    This function transforms a batch of example in tensors
    Each example looks like
    {'src': {'src': ..., 'feats': [...], 'src_ids': ...},
     'tgt': {'tgt': ..., 'tgt_ids': ...},
     'src_original': ['tok1', ...'tokn'],
     'tgt_original': ['tok1', ...'tokm'],
     'cid': corpus id
     'cid_line_number' : corpus id line number
     'ind_in_bucket': index in bucket
     'align': ...,
    }
    Returns  Dict of batch Tensors
        {'src': [seqlen, batchsize, n_feats+1],
         'tgt' : [seqlen, batchsize, n_feats=1],
         'cid': [batchsize],
         'cid_line_number' : [batchsize],
         'ind_in_bucket': [batchsize],
         'srclen': [batchsize],
         'tgtlen': [batchsize],
         'align': alignment sparse tensor
        }
    """
    tensor_batch = {}
    if left_pad:
        tbatchsrc = [
            torch.tensor(ex["src"]["src_ids"], dtype=torch.long, device=device).flip(dims=[0])
            for ex, indice in minibatch
        ]
    else:
        tbatchsrc = [torch.tensor(ex["src"]["src_ids"], dtype=torch.long, device=device) for ex, indice in minibatch]
    padidx = vocabs["src"][vocabs["specials"].get("pad_token", DefaultTokens.PAD)]
    tbatchsrc = pad_sequence(tbatchsrc, batch_first=True, padding_value=padidx)
    """
    This removes some recompiles in torch.dynamo, but slows down and make inference tricky
    tbatchsrc = F.pad(
        tbatchsrc,
        (0, max(0, math.ceil(tbatchsrc.size(1) / 8) * 8 - tbatchsrc.size(1))),
        value=padidx,
    )
    """
    if left_pad:
        tensor_batch["src"] = tbatchsrc.flip(dims=[1])
    else:
        tensor_batch["src"] = tbatchsrc

    tensor_batch["srclen"] = torch.tensor(
        [len(ex["src"]["src_ids"]) for ex, indice in minibatch],
        dtype=torch.long,
        device=device,
    )
    if "prefix_len" in minibatch[0][0]["src"].keys():
        tensor_batch["prefix_len"] = torch.tensor(
            [ex["src"]["prefix_len"] for ex, indice in minibatch],
            dtype=torch.long,
            device=device,
        )
        if left_pad:
            tensor_batch["prefix_len"] += tbatchsrc.eq(padidx).sum(dim=1)
    else:
        tensor_batch["prefix_len"] = None

    if minibatch[0][0].get("tgt", None) is not None:
        if left_pad:
            tbatchtgt = [
                torch.tensor(ex["tgt"]["tgt_ids"], dtype=torch.long, device=device).flip(dims=[0])
                for ex, indice in minibatch
            ]
        else:
            tbatchtgt = [
                torch.tensor(ex["tgt"]["tgt_ids"], dtype=torch.long, device=device) for ex, indice in minibatch
            ]

        padidx = vocabs["tgt"][vocabs["specials"].get("pad_token", DefaultTokens.PAD)]
        tbatchtgt = pad_sequence(tbatchtgt, batch_first=True, padding_value=padidx)

        tbatchtgtlen = torch.tensor(
            [len(ex["tgt"]["tgt_ids"]) for ex, indice in minibatch],
            dtype=torch.long,
            device=device,
        )
        if left_pad:
            tensor_batch["tgt"] = tbatchtgt.flip(dims=[1])
        else:
            tensor_batch["tgt"] = tbatchtgt
        tensor_batch["tgtlen"] = tbatchtgtlen

    if "align" in minibatch[0][0].keys() and minibatch[0][0]["align"] is not None:
        sparse_idx = []
        for i, (ex, indice) in enumerate(minibatch):
            for src, tgt in parse_align_idx(ex["align"]):
                sparse_idx.append([i, tgt + 1, src])
        tbatchalign = torch.tensor(sparse_idx, dtype=torch.long, device=device)
        tensor_batch["align"] = tbatchalign

    if "src_map" in minibatch[0][0].keys():
        src_vocab_size = max([max(ex["src_map"]) for ex, indice in minibatch]) + 1
        src_map = torch.zeros(
            len(tensor_batch["srclen"]),
            tbatchsrc.size(1),
            src_vocab_size,
            device=device,
        )
        for i, (ex, indice) in enumerate(minibatch):
            for j, t in enumerate(ex["src_map"]):
                src_map[i, j, t] = 1
        tensor_batch["src_map"] = src_map

    if "alignment" in minibatch[0][0].keys():
        alignment = torch.zeros(
            len(tensor_batch["srclen"]),
            tbatchtgt.size(1),
            dtype=torch.long,
            device=device,
        )
        for i, (ex, indice) in enumerate(minibatch):
            alignment[i, : len(ex["alignment"])] = torch.tensor(ex["alignment"], dtype=torch.long, device=device)
        tensor_batch["alignment"] = alignment

    if "images" in minibatch[0][0].keys():
        tensor_batch["images"] = [
            torch.tensor(v, device=device, dtype=torch.float32)
            for ex, indice in minibatch
            for k, v in ex["images"].items()
            # BATCH > 1 not supported yet
            # [
            #     torch.tensor(v, device=device, dtype=torch.float32)
            #     for k, v in ex["images"].items()
            # ]
            # for ex, indice in minibatch
        ]
    else:
        tensor_batch["images"] = None

    tensor_batch["ind_in_bucket"] = [indice for ex, indice in minibatch]

    tensor_batch["cid"] = [ex["cid"] for ex, indice in minibatch]
    tensor_batch["cid_line_number"] = [ex["cid_line_number"] for ex, indice in minibatch]

    if minibatch[0][0]["cid"] != "infer":
        tensor_batch["sco"] = torch.tensor([ex["sco"] for ex, indice in minibatch], device=device)

    tensor_batch["left_pad"] = left_pad

    return tensor_batch



""" Prediction main class """

import os
from eole.constants import DefaultTokens
from eole.utils.alignment import build_align_pharaoh


class PredictionBuilder(object):
    """
    Build a word-based prediction from the batch output
    of predictor and the underlying dictionaries.

    Replacement based on "Addressing the Rare Word
    Problem in Neural Machine Translation" :cite:`Luong2015b`

    Args:
       data ():
       vocabs ():
       n_best (int): number of predictions produced
       replace_unk (bool): replace unknown words using attention
    """

    def __init__(
        self,
        vocabs,
        n_best=1,
        replace_unk=False,
        phrase_table="",
        tgt_eos_idx=None,
        id_tokenization=False,
    ):
        self.vocabs = vocabs
        self.n_best = n_best
        self.replace_unk = replace_unk
        self.phrase_table_dict = {}
        self.tgt_eos_idx = tgt_eos_idx  # List of IDs here
        self.id_tokenization = id_tokenization
        if phrase_table != "" and os.path.exists(phrase_table):
            with open(phrase_table) as phrase_table_fd:
                for line in phrase_table_fd:
                    phrase_src, phrase_trg = line.rstrip("\n").split(DefaultTokens.PHRASE_TABLE_SEPARATOR)
                    self.phrase_table_dict[phrase_src] = phrase_trg

    def _build_target_tokens(self, src, srclen, pred, attn, voc):
        if pred is None:
            return []
        pred_list = pred.tolist()
        if pred_list[-1] in self.tgt_eos_idx:
            pred_list = pred_list[:-1]
        if self.id_tokenization:
            tokens = pred_list
        else:
            tokens = [voc[tok] for tok in pred_list]

        # TODO: either support this properly or remove?
        if not self.id_tokenization:
            if self.replace_unk and attn is not None and src is not None:
                for i in range(len(tokens)):
                    if tokens[i] == DefaultTokens.UNK:
                        _, max_index = attn[i][:srclen].max(0)
                        src_tok = self.vocabs["src"].ids_to_tokens[src[max_index.item()]]
                        tokens[i] = src_tok
                        if self.phrase_table_dict:
                            if src_tok in self.phrase_table_dict:
                                tokens[i] = self.phrase_table_dict[src_tok]
        return tokens

    def from_batch(self, prediction_batch):
        batch = prediction_batch["batch"]
        assert len(prediction_batch["gold_score"]) == len(prediction_batch["predictions"])
        batch_size = len(batch["srclen"])

        preds, pred_score, estim, attn, align, gold_score, ind = (
            prediction_batch["predictions"],
            prediction_batch["scores"],
            prediction_batch["estim"],
            prediction_batch["attention"],
            prediction_batch["alignment"],
            prediction_batch["gold_score"],
            batch["ind_in_bucket"],
        )

        if not any(align):  # when align is a empty nested list
            align = [None] * batch_size

        src = batch["src"]
        srclen = batch["srclen"][:]
        if "tgt" in batch.keys():
            tgt = batch["tgt"]
        else:
            tgt = None

        predictions = []
        voc_tgt = self.vocabs["tgt"].ids_to_tokens

        # These comp lists are costy but less than for loops
        for b in range(batch_size):
            pred_sents = [
                self._build_target_tokens(
                    src[b, :] if src is not None else None,
                    srclen[b],
                    preds[b][n] if len(preds[b]) > 0 else None,
                    align[b][n] if align[b] is not None else attn[b][n] if len(attn[b]) > 0 else None,
                    voc_tgt,
                )
                for n in range(self.n_best)
            ]

            gold_sent = None
            if tgt is not None:
                gold_sent = self._build_target_tokens(
                    src[b, :] if src is not None else None,
                    srclen[b],
                    tgt[b, 1:] if tgt is not None else None,
                    None,
                    voc_tgt,
                )

            prediction = Prediction(
                src[b, :] if src is not None else None,
                srclen[b],
                pred_sents,
                attn[b],
                pred_score[b],
                estim[b],
                gold_sent,
                gold_score[b],
                align[b],
                ind[b],
            )
            predictions.append(prediction)

        return predictions


class Prediction(object):
    """Container for a predicted sentence.

    Attributes:
        src (LongTensor): Source word IDs.
        srclen (List[int]): Source lengths.
        pred_sents (List[List[str]]): Words from the n-best predictions.
        pred_scores (List[List[float]]): Log-probs of n-best predictions.
        attns (List[FloatTensor]) : Attention distribution for each
            prediction.
        gold_sent (List[str]): Words from gold prediction.
        gold_score (List[float]): Log-prob of gold prediction.
        word_aligns (List[FloatTensor]): Words Alignment distribution for
            each prediction.
    """

    __slots__ = [
        "src",
        "srclen",
        "pred_sents",
        "attns",
        "pred_scores",
        "estim",
        "gold_sent",
        "gold_score",
        "word_aligns",
        "ind_in_bucket",
    ]

    def __init__(
        self,
        src,
        srclen,
        pred_sents,
        attn,
        pred_scores,
        estim,
        tgt_sent,
        gold_score,
        word_aligns,
        ind_in_bucket,
    ):
        self.src = src
        self.srclen = srclen
        self.pred_sents = pred_sents
        self.attns = attn
        self.pred_scores = pred_scores
        self.estim = estim
        self.gold_sent = tgt_sent
        self.gold_score = gold_score
        self.word_aligns = word_aligns
        self.ind_in_bucket = ind_in_bucket

    def log(self, sent_number, src_raw=""):
        """
        Log prediction.
        """

        msg = ["\nSENT {}: {}\n".format(sent_number, src_raw)]

        best_pred = self.pred_sents[0]
        best_score = self.pred_scores[0]
        best_estim = self.estim[0]
        pred_sent = " ".join([str(x) for x in best_pred])  # this will display IDs for id_tokenize case
        msg.append("PRED {}: {}\n".format(sent_number, pred_sent))
        msg.append("PRED SCORE: {:.4f}\n".format(best_score))
        msg.append("ESTIM SCORE: {:.4f}\n".format(best_estim))

        if self.word_aligns is not None:
            pred_align = self.word_aligns[0]
            pred_align_pharaoh, _ = build_align_pharaoh(pred_align)
            pred_align_sent = " ".join(pred_align_pharaoh)
            msg.append("ALIGN: {}\n".format(pred_align_sent))

        if self.gold_sent is not None:
            tgt_sent = " ".join(self.gold_sent)
            msg.append("GOLD {}: {}\n".format(sent_number, tgt_sent))
            msg.append(("GOLD SCORE: {:.4f}\n".format(self.gold_score)))
        if len(self.pred_sents) > 1:
            msg.append("\nBEST HYP:\n")
            for score, sent in zip(self.pred_scores, self.pred_sents):
                msg.append("[{:.4f}] {}\n".format(score, sent))

        return "".join(msg)



"""
    Code taken and adapted from https://github.com/FranxYao/chain-of-thought-hub
"""

import json
import yaml
import os
import time
from argparse import ArgumentParser
from eole.utils.logging import init_logger
from eole.inference_engine import InferenceEnginePY
from eole.utils.misc import use_gpu, set_random_seed

from eole.bin import BaseBin, register_bin
from eole.config.cli import add_model
from eole.config import get_non_default_values
from eole.config.run import PredictConfig
from eole import ROOT_DIR


TASKS = [
    "abstract_algebra",
    "anatomy",
    "astronomy",
    "business_ethics",
    "clinical_knowledge",
    "college_biology",
    "college_chemistry",
    "college_computer_science",
    "college_mathematics",
    "college_medicine",
    "college_physics",
    "computer_security",
    "conceptual_physics",
    "econometrics",
    "electrical_engineering",
    "elementary_mathematics",
    "formal_logic",
    "global_facts",
    "high_school_biology",
    "high_school_chemistry",
    "high_school_computer_science",
    "high_school_european_history",
    "high_school_geography",
    "high_school_government_and_politics",
    "high_school_macroeconomics",
    "high_school_mathematics",
    "high_school_microeconomics",
    "high_school_physics",
    "high_school_psychology",
    "high_school_statistics",
    "high_school_us_history",
    "high_school_world_history",
    "human_aging",
    "human_sexuality",
    "international_law",
    "jurisprudence",
    "logical_fallacies",
    "machine_learning",
    "management",
    "marketing",
    "medical_genetics",
    "miscellaneous",
    "moral_disputes",
    "moral_scenarios",
    "nutrition",
    "philosophy",
    "prehistory",
    "professional_accounting",
    "professional_law",
    "professional_medicine",
    "professional_psychology",
    "public_relations",
    "security_studies",
    "sociology",
    "us_foreign_policy",
    "virology",
    "world_religions",
]

choices = ["A", "B", "C", "D"]


def compute_metric(run_results):
    total_acc = 0
    total_num = 0
    for task in run_results:
        acc = 0
        pred_answers = run_results[task]["pred_answers"]
        gold_answers = run_results[task]["gold_answers"]
        for pred, gold in zip(pred_answers, gold_answers):
            if pred == gold:
                acc += 1
        print("ACC-%s: %.4f" % (task, acc / len(gold_answers)))
        total_acc += acc
        total_num += len(gold_answers)
        run_results[task]["metrics"] = {"acc": acc}
    acc_all = total_acc / total_num
    print("ACC-all: %.4f" % (acc_all))
    run_results["metrics"] = {"acc": acc_all}
    return run_results


def format_subject(subject):
    return subject.replace("_", " ")


def format_example(df, idx, include_answer=True):
    prompt = df.iloc[idx, 0]
    k = df.shape[1] - 2
    for j in range(k):
        prompt += "\n{}. {}".format(choices[j], df.iloc[idx, j + 1])
    prompt += "\nAnswer:"
    if include_answer:
        prompt += " {}\n\n".format(df.iloc[idx, k + 1])
    return prompt


def gen_prompt(train_df, subject, k=-1):
    prompt = "The following are multiple choice questions (with answers) about {}.\n\n".format(format_subject(subject))
    if k == -1:
        k = train_df.shape[0]
    for i in range(k):
        prompt += format_example(train_df, i)
    return prompt


# def custom_stopping_criteria(input_ids, score, **kwargs):
#     stop_ids = [29871, 13, 13] # \n\n
#     return input_ids[-len(stop_ids)]


def evaluate(args, data_dir):
    import pandas as pd

    logger = init_logger(args.log_file)
    set_random_seed(args.seed, use_gpu(args))

    run_results = {}
    dir_name = args.model_path[0]

    # Save results in the model dir
    output_filename = os.path.join(dir_name, "mmlu_results.json")

    # Build the translator (along with the model)
    engine = InferenceEnginePY(args)

    ntrain = 5  # nshots from dev

    start_time = time.time()
    for task in TASKS:
        logger.info("Testing %s ..." % task)
        records = []
        src = []
        dev_df = pd.read_csv(os.path.join(data_dir, "dev", task + "_dev.csv"), header=None)[:ntrain]
        test_df = pd.read_csv(os.path.join(data_dir, "test", task + "_test.csv"), header=None)
        for i in range(test_df.shape[0]):
            # get prompt and make sure it fits
            k = ntrain
            prompt_end = format_example(test_df, i, include_answer=False)
            train_prompt = gen_prompt(dev_df, task, k)
            prompt = train_prompt + prompt_end

            while len(prompt.split(" ")) > 768:
                prompt_split = prompt.split("\n\n")
                prompt_split.pop(1)
                prompt = "\n\n".join(prompt_split)

            label = test_df.iloc[i, test_df.shape[1] - 1]
            records.append({"prompt": prompt, "answer": label})
            if "onmt_tokenize" in args.transforms:
                src.append(prompt.replace("\n", "｟newline｠"))
            else:
                src.append(prompt)

        scores, _, preds = engine.infer_list(src)

        pred_answers = [x.lstrip() for sublist in preds for x in sublist]  # flatten the list of list

        gold_answers = [record["answer"] for record in records]
        run_results[task] = {"pred_answers": pred_answers, "gold_answers": gold_answers}

    engine.terminate()

    run_results = compute_metric(run_results)

    with open(output_filename, "w", encoding="utf-8") as f:
        json.dump(run_results, f, ensure_ascii=False, indent=2)

    end_time = time.time()
    logger.info("total run time %.2f" % (end_time - start_time))


@register_bin(name="run_mmlu")
class RunMMLU(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "-config",
            "--config",
            "-c",
            required=False,
            help="Path of main YAML config file.",
        )
        # TODO: we might want to retrieve transparently from HF at some point
        parser.add_argument(
            "-data_dir",
            "--data_dir",
            "-d",
            required=False,
            help="Path to the MMLU data root.",
            default=os.path.join(os.path.dirname(ROOT_DIR), "recipes", "mmlu", "data"),
        )
        add_model(parser, PredictConfig)

    def run(args):
        # same as in eole.bin.translate, should be improved/factorized
        if args.config is not None:
            with open(args.config) as f:
                config = yaml.safe_load(os.path.expandvars(f.read()))
        else:
            config = {}
        _parser = ArgumentParser()
        add_model(_parser, PredictConfig)
        defaults = vars(_parser.parse_args([]))
        stuff_to_update = get_non_default_values(args, defaults)
        config.update(stuff_to_update)

        # pop extra fields added by argparse, not supported in pydantic configs
        if "bin" in config.keys():
            config.pop("bin")
        if "sub_bin" in config.keys():
            config.pop("sub_bin")
        if "config" in config.keys():
            config.pop("config")

        # not supported in PredictConfig schema
        data_dir = config.pop("data_dir")

        # models retrieved from HF might have some default inference settings,
        # let's manually override a few things for the MMLU context
        config.update(
            {
                "top_p": 0.0,
                "top_k": 0,
                "beam_size": 1,
                "n_best": 1,
                "max_length": 1,
                "src": "dummy",
            }
        )

        config = PredictConfig(**config)

        evaluate(config, data_dir)



""" Report manager utility """

import time
from datetime import datetime

import eole

from eole.utils.logging import logger


def build_report_manager(config, gpu_rank):
    if config.tensorboard and gpu_rank <= 0:
        from torch.utils.tensorboard import SummaryWriter

        if getattr(config, "tensorboard_log_dir_dated", None) is None:
            config.tensorboard_log_dir_dated = config.tensorboard_log_dir + datetime.now().strftime("/%b-%d_%H-%M-%S")
        writer = SummaryWriter(config.tensorboard_log_dir_dated, comment="Unmt")
    else:
        writer = None

    report_mgr = ReportMgr(config.report_every, start_time=-1, tensorboard_writer=writer)
    return report_mgr


class ReportMgrBase(object):
    """
    Report Manager Base class
    Inherited classes should override:
        * `_report_training`
        * `_report_step`
    """

    def __init__(self, report_every, start_time=-1.0):
        """
        Args:
            report_every(int): Report status every this many sentences
            start_time(float): manually set report start time. Negative values
                means that you will need to set it later or use `start()`
        """
        self.report_every = report_every
        self.start_time = start_time

    def start(self):
        self.start_time = time.time()

    def log(self, *args, **kwargs):
        logger.info(*args, **kwargs)

    def report_training(self, step, num_steps, learning_rate, patience, report_stats, multigpu=False):
        """
        This is the user-defined batch-level traing progress
        report function.
        Args:
            step(int): current step count.
            num_steps(int): total number of batches.
            learning_rate(float): current learning rate.
            report_stats(Statistics): old Statistics instance.
        Returns:
            report_stats(Statistics): updated Statistics instance.
        """
        if self.start_time < 0:
            raise ValueError(
                """ReportMgr needs to be started
                                (set 'start_time' or use 'start()'"""
            )

        if step % self.report_every == 0:
            if multigpu:
                report_stats = eole.utils.Statistics.all_gather_stats(report_stats)
            self._report_training(step, num_steps, learning_rate, patience, report_stats)
            return eole.utils.Statistics()
        else:
            return report_stats

    def _report_training(self, *args, **kwargs):
        """To be overridden"""
        raise NotImplementedError()

    def report_step(self, lr, patience, step, train_stats=None, valid_stats=None):
        """
        Report stats of a step

        Args:
            lr(float): current learning rate
            patience(int): current patience
            step(int): current step
            train_stats(Statistics): training stats
            valid_stats(Statistics): validation stats
        """
        self._report_step(lr, patience, step, valid_stats=valid_stats, train_stats=train_stats)

    def _report_step(self, *args, **kwargs):
        raise NotImplementedError()


class ReportMgr(ReportMgrBase):
    def __init__(self, report_every, start_time=-1.0, tensorboard_writer=None):
        """
        A report manager that writes statistics on standard output as well as
        (optionally) TensorBoard

        Args:
            report_every(int): Report status every this many sentences
            tensorboard_writer(:obj:`tensorboard.SummaryWriter`):
                The TensorBoard Summary writer to use or None
        """
        super(ReportMgr, self).__init__(report_every, start_time)
        self.tensorboard_writer = tensorboard_writer

    def maybe_log_tensorboard(self, stats, prefix, learning_rate, patience, step):
        if self.tensorboard_writer is not None:
            stats.log_tensorboard(prefix, self.tensorboard_writer, learning_rate, patience, step)

    def _report_training(self, step, num_steps, learning_rate, patience, report_stats):
        """
        See base class method `ReportMgrBase.report_training`.
        """
        report_stats.output(step, num_steps, learning_rate, self.start_time)
        self.maybe_log_tensorboard(report_stats, "progress", learning_rate, patience, step)
        report_stats = eole.utils.Statistics()

        return report_stats

    def _report_step(self, lr, patience, step, valid_stats=None, train_stats=None):
        """
        See base class method `ReportMgrBase.report_step`.
        """
        if train_stats is not None:
            self.log("Train perplexity: %g" % train_stats.ppl())
            self.log("Train accuracy: %g" % train_stats.accuracy())
            self.log("Sentences processed: %g" % train_stats.n_sents)
            self.log(train_stats.data_stats)
            self.log(
                "Average bsz: %4.0f/%4.0f/%2.0f"
                % (
                    train_stats.n_src_tokens / train_stats.n_batchs,
                    train_stats.n_tokens / train_stats.n_batchs,
                    train_stats.n_sents / train_stats.n_batchs,
                )
            )

            self.maybe_log_tensorboard(train_stats, "train", lr, patience, step)
        if valid_stats is not None:
            self.log("Validation perplexity: %g" % valid_stats.ppl())
            self.log("Validation accuracy: %g" % valid_stats.accuracy())

            self.maybe_log_tensorboard(valid_stats, "valid", lr, patience, step)



from eole.constants import DefaultTokens
from eole.transforms import register_transform
from .transform import Transform, TransformConfig
from pydantic import Field
import copy


class DocifyConfig(TransformConfig):
    doc_length: int | None = Field(default=200, description="Number of tokens per doc.")
    max_context: int | None = Field(default=1, description="Max context segments.")


@register_transform(name="docify")
class DocifyTransform(Transform):
    """
    Convert source and target examples to doc level segments.

    It concatenates segments with a DefaultTokens.SEP
    until it reaches --doc_length tokens

    """

    config_model = DocifyConfig

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        if hasattr(self.config, "num_workers") and hasattr(self.config, "world_size"):
            self.stride = self.config.num_workers * self.config.world_size
        else:
            self.stride = 1
        self.doc_length = self.config.doc_length
        self.max_context = self.config.max_context

    @classmethod
    def get_specials(cls, config):
        """Add newline tag to src and tgt vocabs."""

        src_specials, tgt_specials = [DefaultTokens.SEP], [DefaultTokens.SEP]
        return (src_specials, tgt_specials)

    def warm_up(self, vocabs=None):
        super().warm_up(None)
        if self.stride != 1:
            assert (
                self.stride % (self.max_context + 1) == 0
            ), "(max_context+1) must be a multiple \
                 of num_workers * world_size"

    def batch_apply(self, batch, is_train=False, stats=None, **kwargs):
        """Convert source and target examples to doc level segments."""
        if self.max_context == 0:
            return batch
        trf_batch = []
        doc = {}
        doc["src"] = []
        doc["tgt"] = []
        doc["cid"] = ""
        doc["cid_line_number"] = 0

        for ex, _, cid in batch:
            if ex["tgt"] is not None:
                cur_len = max(len(doc["src"] + ex["src"]), len(doc["tgt"] + ex["tgt"]))

                if len(ex["src"]) == 0 and len(ex["tgt"]) == 0:
                    # doc break we add it, restart new doc
                    trf_batch.append((doc, self, cid))
                    doc = {}
                    doc["src"] = []
                    doc["tgt"] = []
                    doc["cid"] = ex["cid"]
                    doc["cid_line_number"] = ex["cid_line_number"]
                elif cur_len > self.doc_length:
                    if len(doc["src"]) == 0:
                        # case 1st ex is already longer
                        trf_batch.append((ex, self, cid))
                    else:
                        # adding cur ex is too long we add cur doc
                        # and reset doc to cur ex
                        trf_batch.append((doc, self, cid))
                        doc = copy.deepcopy(ex)
                else:
                    if len(doc["src"]) == 0:
                        # we start the new doc with cur ex
                        doc = copy.deepcopy(ex)
                    else:
                        # we cumulate cur ex to cur doc
                        doc["src"] += [DefaultTokens.SEP] + ex["src"]
                        doc["src_original"] += [DefaultTokens.SEP] + ex["src_original"]
                        doc["tgt"] += [DefaultTokens.SEP] + ex["tgt"]
                        doc["tgt_original"] += [DefaultTokens.SEP] + ex["tgt_original"]
                        nb_ctx = doc["src"].count(DefaultTokens.SEP)
                        if nb_ctx >= self.max_context:
                            trf_batch.append((doc, self, cid))
                            doc = {}
                            doc["src"] = []
                            doc["tgt"] = []
                            doc["cid"] = ex["cid"]
                            doc["cid_line_number"] = ex["cid_line_number"]
            else:
                cur_len = len(doc["src"] + ex["src"])
                doc["tgt"] = None
                if len(ex["src"]) == 0:
                    trf_batch.append((doc, self, cid))
                    doc = {}
                    doc["src"] = []
                    doc["indices"] = ex["indices"]
                elif cur_len > self.doc_length:
                    if len(doc["src"]) == 0:
                        trf_batch.append((ex, self, cid))
                    else:
                        trf_batch.append((doc, self, cid))
                        doc = copy.deepcopy(ex)
                else:
                    if len(doc["src"]) == 0:
                        doc = copy.deepcopy(ex)
                    else:
                        doc["src"] += [DefaultTokens.SEP] + ex["src"]
                        doc["src_original"] += [DefaultTokens.SEP] + ex["src_original"]
                        nb_ctx = doc["src"].count(DefaultTokens.SEP)
                        if nb_ctx >= self.max_context:
                            trf_batch.append((doc, self, cid))
                            doc = {}
                            doc["src"] = []
                            doc["cid"] = ex["cid"]
                            doc["cid_line_number"] = ex["cid_line_number"]
        if len(doc["src"]) > 0:
            trf_batch.append((doc, self, cid))
        return trf_batch

    def apply_reverse(self, predicted):
        segments = predicted.split(DefaultTokens.SEP)
        segments = [segment.strip(" ") for segment in segments]
        return segments



from copy import deepcopy
from collections import defaultdict

from eole.config.models import (
    TransformerEncoderModelConfig,
    TransformerModelConfig,
    TransformerLMModelConfig,
    VisionTransformerLMModelConfig,
)
from eole.config import recursive_update_dict
from eole.constants import PositionEncodingType

# Default tensor key mappings, based on Llama
BASE_KEY_MAP = {
    "decoder_layer_prefix": "model.layers.",
    # keys outside of encoder/decoder subsections will be centralized in the first shard
    "tgt_emb.embeddings.weight": "model.embed_tokens.weight",
    "decoder.layer_norm.weight": "model.norm.weight",
    "generator.weight": "lm_head.weight",
    "decoder": {
        ".self_attn.linear_query.": ".self_attn.q_proj.",
        ".self_attn.linear_keys.": ".self_attn.k_proj.",
        ".self_attn.linear_values.": ".self_attn.v_proj.",
        ".self_attn.final_linear.": ".self_attn.o_proj.",
        ".mlp.gate_up_proj.": ".mlp.gate_proj.",
        ".mlp.down_proj.": ".mlp.down_proj.",
        ".mlp.up_proj.": ".mlp.up_proj.",
        ".input_layernorm.": ".input_layernorm.",
        ".post_attention_layernorm.": ".post_attention_layernorm.",
    },
}


# Model-specific overrides for key mappings and configurations
# root keys are weights to be added in the first shard
# encoder/decoder sections are modules of each encoder/decoder layer
MODEL_OVERRIDES = {
    "LlamaForCausalLM": {},  # default
    "MistralForCausalLM": {},
    "Qwen2ForCausalLM": {
        "config": {
            "add_qkvbias": True,
            "add_final_linear_bias": False,
        }
    },
    "Qwen3ForCausalLM": {
        "decoder": {
            ".self_attn.q_norm.": ".self_attn.q_norm.",
            ".self_attn.k_norm.": ".self_attn.k_norm.",
        },
        "config": {
            "decoder": {
                "query_norm": True,
                "key_norm": True,
            }
        },
    },
    "Qwen3MoeForCausalLM": {
        "decoder": {
            ".self_attn.q_norm.": ".self_attn.q_norm.",
            ".self_attn.k_norm.": ".self_attn.k_norm.",
            ".mlp.gate.": ".mlp.gate.",
            **{f".mlp.experts.{i}.gate_up_proj.": f".mlp.experts.{i}.gate_proj." for i in range(128)},
            **{f".mlp.experts.{i}.up_proj.": f".mlp.experts.{i}.up_proj." for i in range(128)},
            **{f".mlp.experts.{i}.down_proj.": f".mlp.experts.{i}.down_proj." for i in range(128)},
        },
        "config": {
            "decoder": {
                "query_norm": True,
                "key_norm": True,
            }
        },
    },
    "Gemma2ForCausalLM": {
        "decoder": {
            ".pre_feedforward_layernorm.": ".pre_feedforward_layernorm.",
            ".post_feedforward_layernorm.": ".post_feedforward_layernorm.",
        },
        "config": {
            "share_decoder_embeddings": True,
            "ffn_layernorm": True,
            "embeddings": {
                "normalize": True,
            },
        },
    },
    "MixtralForCausalLM": {
        "decoder": {
            ".mlp.gate.": ".block_sparse_moe.gate.",
            **{
                f".mlp.experts.{i}.{attr}": f".block_sparse_moe.experts.{i}.w{j+1}."
                for i in range(8)
                for j, attr in enumerate(["gate_up_proj.", "down_proj.", "up_proj."])
            },
        },
        "config": {
            "decoder": {"moe_transformer_ff": 14336},
            "moe_softmax_after": True,
            "shared_layer_norm": True,
        },
    },
    "PhiForCausalLM": {
        "decoder.layer_norm.weight": "model.final_layernorm.weight",
        "decoder.layer_norm.bias": "model.final_layernorm.bias",
        "generator.bias": "lm_head.bias",
        "decoder": {
            ".self_attn.final_linear.": ".self_attn.dense.",
            ".mlp.gate_up_proj.": ".mlp.fc1.",
            ".mlp.down_proj.": ".mlp.fc2.",
            ".input_layernorm.": (".input_layernorm.", ""),
        },
        "config": {
            "parallel_residual": True,
            "shared_layer_norm": True,
            "add_qkvbias": True,
            "add_final_linear_bias": True,
            "add_ffnbias": True,
        },
    },
    "Phi3ForCausalLM": {
        "decoder": {
            ".self_attn.linear_query.": (".self_attn.qkv_proj.", "[:hidden_size, :]"),
            ".self_attn.linear_keys.": (
                ".self_attn.qkv_proj.",
                "[hidden_size:2*hidden_size, :]",
            ),
            ".self_attn.linear_values.": (".self_attn.qkv_proj.", "[-hidden_size:, :]"),
            ".mlp.gate_up_proj.": (".mlp.gate_up_proj.", "[:transformer_ff, :]"),
            ".mlp.up_proj.": (".mlp.gate_up_proj.", "[transformer_ff:, :]"),
        }
    },
    "PhiMoEForCausalLM": {
        "decoder": {
            ".mlp.gate.": ".block_sparse_moe.gate.",
            **{
                f".mlp.experts.{i}.{attr}": f".block_sparse_moe.experts.{i}.w{j+1}."
                for i in range(16)
                for j, attr in enumerate(["gate_up_proj.", "down_proj.", "up_proj."])
            },
        },
        "config": {
            "decoder": {"moe_transformer_ff": 6400},
            "add_qkvbias": True,
            "add_final_linear_bias": True,
        },
    },
    "HunYuanDenseV1ForCausalLM": {
        "decoder": {
            ".self_attn.q_norm.": ".self_attn.query_layernorm.",
            ".self_attn.k_norm.": ".self_attn.key_layernorm.",
        },
        "config": {
            "query_norm": True,
            "key_norm": True,
            "qk_norm_post_rope": True,
        },
    },
    "GPT2LMHeadModel": {
        "decoder_layer_prefix": "h.",
        "decoder.layer_norm.weight": "ln_f.weight",
        "decoder.layer_norm.bias": "ln_f.bias",
        "tgt_emb.pe.weight": "wpe.weight",
        "tgt_emb.embeddings.weight": "wte.weight",
        "generator.weight": "wte.weight",  # shared with embeddings
        "decoder": {
            ".self_attn.linear_query.": (".attn.c_attn.", ".t()[:hidden_size, ...]"),
            ".self_attn.linear_keys.": (
                ".attn.c_attn.",
                ".t()[hidden_size:2*hidden_size, ...]",
            ),
            ".self_attn.linear_values.": (".attn.c_attn.", ".t()[-hidden_size:, ...]"),
            ".self_attn.final_linear.": (".attn.c_proj.", ".t()"),
            ".mlp.gate_up_proj.": (".mlp.c_fc.", ".t()"),
            ".mlp.down_proj.": (".mlp.c_proj.", ".t()"),
            ".input_layernorm.": ".ln_1.",
            ".post_attention_layernorm.": ".ln_2.",
        },
        "config": {
            "parallel_residual": False,
            "shared_layer_norm": True,
            "add_qkvbias": True,
            "add_final_linear_bias": True,
            "add_ffnbias": True,
            "embeddings": {
                "position_encoding_type": "Learned",
                "n_positions": 1024,
            },
            "left_pad": False,
        },
    },
    "XLMRobertaXLForMaskedLM": {
        "encoder_layer_prefix": "roberta.encoder.layer.",
        "encoder.layer_norm.weight": "roberta.encoder.LayerNorm.weight",
        "encoder.layer_norm.bias": "roberta.encoder.LayerNorm.bias",
        "src_emb.embeddings.weight": "roberta.embeddings.word_embeddings.weight",
        "src_emb.pe.weight": "roberta.embeddings.position_embeddings.weight",
        "encoder": {
            ".self_attn.linear_query.": ".attention.self.query.",
            ".self_attn.linear_keys.": ".attention.self.key.",
            ".self_attn.linear_values.": ".attention.self.value.",
            ".self_attn.final_linear.": ".attention.output.dense.",
            ".mlp.gate_up_proj.": ".intermediate.dense.",
            ".mlp.down_proj.": ".output.dense.",
            ".input_layernorm.": ".attention.self_attn_layer_norm.",
            ".post_attention_layernorm.": ".LayerNorm.",
        },
        "config": {
            "add_qkvbias": True,
            "add_final_linear_bias": True,
            "add_ffnbias": True,
            "embeddings": {
                "position_encoding_type": "Learned",
                "n_positions": 514,
                "position_shift": 2,
            },
            "left_pad": False,
        },
    },
    "LlavaForConditionalGeneration": {
        "decoder_layer_prefix": "language_model.model.layers.",
        "tgt_emb.embeddings.weight": "language_model.model.embed_tokens.weight",
        "decoder.layer_norm.weight": "language_model.model.norm.weight",
        "generator.weight": "language_model.lm_head.weight",
        "encoder.patch_conv.weight": "vision_tower.patch_conv.weight",
        "encoder.ln_pre.weight": "vision_tower.ln_pre.weight",
        "encoder_layer_prefix": "vision_tower.transformer.layers.",
        "encoder": {
            "layers": 24,
            ".self_attn.linear_query.": ".attention.q_proj.",
            ".self_attn.linear_keys.": ".attention.k_proj.",
            ".self_attn.linear_values.": ".attention.v_proj.",
            ".self_attn.final_linear.": ".attention.o_proj.",
            ".mlp.gate_up_proj.": ".feed_forward.gate_proj.",
            ".mlp.down_proj.": ".feed_forward.down_proj.",
            ".mlp.up_proj.": ".feed_forward.up_proj.",
            ".input_layernorm.": ".attention_norm.",
            ".post_attention_layernorm.": ".ffn_norm.",
        },
        # vision_adapter
        "adapter.w_in.weight": "multi_modal_projector.linear_1.weight",
        "adapter.w_in.bias": "multi_modal_projector.linear_1.bias",
        "adapter.w_out.weight": "multi_modal_projector.linear_2.weight",
        "adapter.w_out.bias": "multi_modal_projector.linear_2.bias",
        "config": {
            "encoder": {
                "num_channels": 3,
                "image_token_id": 10,
                "layernorm_pre": True,
                "patch_conv_bias": False,
            },
        },
    },
    "Mistral3ForConditionalGeneration": {
        "decoder_layer_prefix": "language_model.model.layers.",
        "tgt_emb.embeddings.weight": "language_model.model.embed_tokens.weight",
        "decoder.layer_norm.weight": "language_model.model.norm.weight",
        "generator.weight": "language_model.lm_head.weight",
        "encoder.patch_conv.weight": "vision_tower.patch_conv.weight",
        "encoder.ln_pre.weight": "vision_tower.ln_pre.weight",
        # vision_tower
        "encoder_layer_prefix": "vision_tower.transformer.layers.",
        "encoder": {
            "layers": 24,
            ".self_attn.linear_query.": ".attention.q_proj.",
            ".self_attn.linear_keys.": ".attention.k_proj.",
            ".self_attn.linear_values.": ".attention.v_proj.",
            ".self_attn.final_linear.": ".attention.o_proj.",
            ".mlp.gate_up_proj.": ".feed_forward.gate_proj.",
            ".mlp.down_proj.": ".feed_forward.down_proj.",
            ".mlp.up_proj.": ".feed_forward.up_proj.",
            ".input_layernorm.": ".attention_norm.",
            ".post_attention_layernorm.": ".ffn_norm.",
        },
        # vision_adapter
        "adapter.w_in.weight": "multi_modal_projector.linear_1.weight",
        "adapter.w_out.weight": "multi_modal_projector.linear_2.weight",
        "adapter.layernorm.weight": "multi_modal_projector.norm.weight",
        "adapter.patch_merger.merging_layer.weight": "multi_modal_projector.patch_merger.merging_layer.weight",
        "config": {
            "encoder": {
                "num_channels": 3,
                "image_token_id": 10,
                "layernorm_pre": True,
                "patch_conv_bias": False,
            },
        },
    },
    "Gemma3ForConditionalGeneration": {
        "decoder_layer_prefix": "language_model.model.layers.",
        "tgt_emb.embeddings.weight": "language_model.model.embed_tokens.weight",
        "decoder.layer_norm.weight": "language_model.model.norm.weight",
        # "generator.weight": "language_model.lm_head.weight", # probably shared with embeddings
        # decoder layer modules
        "decoder": {
            ".self_attn.q_norm.": ".self_attn.q_norm.",
            ".self_attn.k_norm.": ".self_attn.k_norm.",
            ".pre_feedforward_layernorm.": ".pre_feedforward_layernorm.",
            ".post_feedforward_layernorm.": ".post_feedforward_layernorm.",
        },
        "encoder_layer_prefix": "vision_tower.vision_model.encoder.layers.",
        "encoder.patch_conv.weight": "vision_tower.vision_model.embeddings.patch_embedding.weight",
        "encoder.patch_conv.bias": "vision_tower.vision_model.embeddings.patch_embedding.bias",
        "encoder.post_layernorm.weight": "vision_tower.vision_model.post_layernorm.weight",
        "encoder.post_layernorm.bias": "vision_tower.vision_model.post_layernorm.bias",
        "encoder.position_embeddings.weight": "vision_tower.vision_model.embeddings.position_embedding.weight",
        # "encoder.ln_pre.weight": "vision_tower.ln_pre.weight", # no ln_pre in Gemma3
        # encoder layers modules
        "encoder": {
            ".self_attn.linear_query.": ".self_attn.q_proj.",
            ".self_attn.linear_keys.": ".self_attn.k_proj.",
            ".self_attn.linear_values.": ".self_attn.v_proj.",
            ".self_attn.final_linear.": ".self_attn.out_proj.",
            ".mlp.gate_up_proj.": ".mlp.fc1.",
            ".mlp.down_proj.": ".mlp.fc2.",
            ".input_layernorm.": ".layer_norm1.",
            ".post_attention_layernorm.": ".layer_norm2.",
        },
        "adapter.w_in.weight": ("multi_modal_projector.mm_input_projection_weight", ".t()"),
        "adapter.norm.weight": "multi_modal_projector.mm_soft_emb_norm.weight",
        "config": {
            "share_decoder_embeddings": True,
            "ffn_layernorm": True,
            "embeddings": {
                "normalize": True,
            },
            "adapter": "gemma3",
            "decoder": {
                "query_norm": True,
                "key_norm": True,
                "rope_config": {
                    "rotary_theta": 1000000,
                    "scaling_type": "gemma3",
                    "rotary_theta_local": 10000,
                    "interleave_local": 6,
                    "tmax_index": 1,
                },
            },
            "encoder": {
                "mlp_activation_fn": "gelu-tanh",
                "position_encoding_type": PositionEncodingType.Learned,
                "layer_norm": "standard",
                "add_ffnbias": True,
                "add_final_linear_bias": True,
                "add_qkvbias": True,
                "layernorm_pre": False,
                "layernorm_post": True,
                "patch_conv_bias": True,
            },
        },
    },
    "M2M100ForConditionalGeneration": {
        "decoder_layer_prefix": "model.decoder.layers.",
        "src_emb.embeddings.weight": "model.encoder.embed_tokens.weight",
        "tgt_emb.embeddings.weight": "model.decoder.embed_tokens.weight",
        "decoder.layer_norm.weight": "model.decoder.layer_norm.weight",
        "decoder.layer_norm.bias": "model.decoder.layer_norm.bias",
        "decoder": {
            ".self_attn.linear_query.": ".self_attn.q_proj.",
            ".self_attn.linear_keys.": ".self_attn.k_proj.",
            ".self_attn.linear_values.": ".self_attn.v_proj.",
            ".self_attn.final_linear.": ".self_attn.out_proj.",
            ".precontext_layernorm.": ".encoder_attn_layer_norm.",
            ".context_attn.linear_query.": ".encoder_attn.q_proj.",
            ".context_attn.linear_keys.": ".encoder_attn.k_proj.",
            ".context_attn.linear_values.": ".encoder_attn.v_proj.",
            ".context_attn.final_linear.": ".encoder_attn.out_proj.",
            ".mlp.gate_up_proj.": ".fc1.",
            ".mlp.down_proj.": ".fc2.",
            ".input_layernorm.": ".self_attn_layer_norm.",
            ".post_attention_layernorm.": ".final_layer_norm.",
        },
        "encoder_layer_prefix": "model.encoder.layers.",
        "encoder.layer_norm.weight": "model.encoder.layer_norm.weight",
        "encoder.layer_norm.bias": "model.encoder.layer_norm.bias",
        "encoder": {
            ".self_attn.linear_query.": ".self_attn.q_proj.",
            ".self_attn.linear_keys.": ".self_attn.k_proj.",
            ".self_attn.linear_values.": ".self_attn.v_proj.",
            ".self_attn.final_linear.": ".self_attn.out_proj.",
            ".mlp.gate_up_proj.": ".fc1.",
            ".mlp.down_proj.": ".fc2.",
            ".input_layernorm.": ".self_attn_layer_norm.",
            ".post_attention_layernorm.": ".final_layer_norm.",
        },
        "config": {
            "parallel_residual": False,
            "add_qkvbias": True,
            "add_final_linear_bias": True,
            "add_ffnbias": True,
            "embeddings": {
                "position_encoding_type": "SinusoidalConcat",
                "n_positions": 1024,
            },
            "left_pad": False,
            "share_decoder_embeddings": True,
        },
    },
    "DeepseekOCRForCausalLM": {
        "decoder": {
            ".mlp.gate.": ".mlp.gate.",
            **{
                f".mlp.experts.{i}.{k}": f".mlp.experts.{i}.{v}"
                for i in range(64)
                for k, v in {"gate_up_proj.": "gate_proj.", "down_proj.": "down_proj.", "up_proj.": "up_proj."}.items()
            },
            **{
                f".mlp.shared_experts.{k}": f".mlp.shared_experts.{v}"
                for k, v in {"gate_up_proj.": "gate_proj.", "down_proj.": "down_proj.", "up_proj.": "up_proj."}.items()
            },
        },
        "encoder.patch_conv.weight": "model.vision_model.embeddings.patch_embedding.weight",
        "encoder.position_embeddings.weight": "model.vision_model.embeddings.position_embedding.weight",
        "encoder.class_embedding.weight": "model.vision_model.embeddings.class_embedding",
        "encoder.ln_pre.weight": "model.vision_model.pre_layrnorm.weight",
        "encoder.ln_pre.bias": "model.vision_model.pre_layrnorm.bias",
        # vision_tower
        "encoder_layer_prefix": "model.vision_model.transformer.layers.",
        "encoder": {
            "layers": 24,
            ".self_attn.linear_query.weight": (".self_attn.qkv_proj.weight", "[:hidden_size, :]"),
            ".self_attn.linear_keys.weight": (".self_attn.qkv_proj.weight", "[hidden_size:2*hidden_size, :]"),
            ".self_attn.linear_values.weight": (".self_attn.qkv_proj.weight", "[-hidden_size:, :]"),
            ".self_attn.linear_query.bias": (".self_attn.qkv_proj.bias", "[:hidden_size]"),
            ".self_attn.linear_keys.bias": (".self_attn.qkv_proj.bias", "[hidden_size:2*hidden_size]"),
            ".self_attn.linear_values.bias": (".self_attn.qkv_proj.bias", "[-hidden_size:]"),
            ".self_attn.final_linear.": ".self_attn.out_proj.",
            ".mlp.gate_up_proj.": ".mlp.fc1.",
            ".mlp.down_proj.": ".mlp.fc2.",
            ".input_layernorm.": ".layer_norm1.",
            ".post_attention_layernorm.": ".layer_norm2.",
        },
        # vision_adapter
        "adapter.w_in.weight": "model.projector.layers.weight",
        "adapter.w_in.bias": "model.projector.layers.bias",
        # Misc deepseek ocr
        "image_newline": "model.image_newline",
        "view_separator": "model.view_seperator",
        # sam_model
        "encoder.sam.patch_embed.proj.weight": "model.sam_model.patch_embed.proj.weight",
        "encoder.sam.patch_embed.proj.bias": "model.sam_model.patch_embed.proj.bias",
        "encoder.sam.pos_embed": "model.sam_model.pos_embed",
        "encoder.sam.neck.0.weight": "model.sam_model.neck.0.weight",
        "encoder.sam.neck.1.weight": "model.sam_model.neck.1.weight",
        "encoder.sam.neck.1.bias": "model.sam_model.neck.1.bias",
        "encoder.sam.neck.2.weight": "model.sam_model.neck.2.weight",
        "encoder.sam.neck.3.weight": "model.sam_model.neck.3.weight",
        "encoder.sam.neck.3.bias": "model.sam_model.neck.3.bias",
        "encoder.sam.net_2.weight": "model.sam_model.net_2.weight",
        "encoder.sam.net_3.weight": "model.sam_model.net_3.weight",
        "encoder_sam_layer_prefix": "model.sam_model.blocks.",
        "encoder.sam": {
            "layers": 12,
            ".attn.proj.": ".attn.proj.",
            ".attn.qkv.": ".attn.qkv.",
            ".attn.rel_pos_h": ".attn.rel_pos_h",
            ".attn.rel_pos_w": ".attn.rel_pos_w",
            ".mlp.lin1.": ".mlp.lin1.",
            ".mlp.lin2.": ".mlp.lin2.",
            ".norm1.": ".norm1.",
            ".norm2.": ".norm2.",
        },
        "config": {
            "adapter": "deepseekocr",
            "adapter_bias": True,
            "decoder": {
                "layer_norm": "rms",
                "norm_eps": 1e-6,
            },
            "encoder": {
                "mlp_activation_fn": "quick_gelu",
                "layer_norm": "standard",
                "norm_eps": 1e-5,
                "position_encoding_type": PositionEncodingType.Learned,
                "add_ffnbias": True,
                "add_final_linear_bias": True,
                "add_qkvbias": True,
                "num_channels": 3,
                "image_token_id": 128815,
                "layernorm_pre": True,
                "patch_conv_bias": False,
                "encoder_sam": True,
                "use_class_embedding": True,
            },
        },
    },
    "HunYuanVLForConditionalGeneration": {
        "decoder": {
            ".self_attn.q_norm.": ".self_attn.query_layernorm.",
            ".self_attn.k_norm.": ".self_attn.key_layernorm.",
        },
        "config": {
            "adapter": "hunyuanocr",
            "decoder": {
                "query_norm": True,
                "key_norm": True,
                "qk_norm_post_rope": True,
            },
            "encoder": {
                "position_encoding_type": PositionEncodingType.Learned,
                "interpolate_mode": "bilinear",
                "n_positions": 16384,
                "layer_norm": "standard",
                "layernorm_pre": False,
                "mlp_activation_fn": "gelu",
                "add_ffnbias": True,
                "add_qkvbias": True,
                "patch_conv_bias": True,
                "image_token_id": 120120,
            },
        },
        "encoder_layer_prefix": "vit.layers.",
        "encoder.patch_conv.weight": "vit.embeddings.patch_embedding.weight",
        "encoder.patch_conv.bias": "vit.embeddings.patch_embedding.bias",
        "encoder.position_embeddings.weight": "vit.embeddings.position_embedding.weight",
        # encoder layers modules
        "encoder": {
            ".self_attn.linear_query.": ".self_attn.q_proj.",
            ".self_attn.linear_keys.": ".self_attn.k_proj.",
            ".self_attn.linear_values.": ".self_attn.v_proj.",
            ".self_attn.final_linear.": ".self_attn.o_proj.",
            ".mlp.gate_up_proj.": ".mlp.dense_h_to_4h.",
            ".mlp.down_proj.": ".mlp.dense_4h_to_h.",
            ".input_layernorm.": ".input_layernorm.",
            ".post_attention_layernorm.": ".post_attention_layernorm.",
        },
        "adapter.proj.0.weight": "vit.perceive.proj.0.weight",
        "adapter.proj.2.weight": "vit.perceive.proj.2.weight",
        "adapter.mlp.weight": "vit.perceive.mlp.weight",
        "adapter.proj.0.bias": "vit.perceive.proj.0.bias",
        "adapter.proj.2.bias": "vit.perceive.proj.2.bias",
        "adapter.mlp.bias": "vit.perceive.mlp.bias",
        "adapter.after_rms.weight": "vit.perceive.after_rms.weight",
        "adapter.before_rms.weight": "vit.perceive.before_rms.weight",
        "adapter.image_begin": "vit.perceive.image_begin",
        "adapter.image_end": "vit.perceive.image_end",
        "adapter.image_newline": "vit.perceive.image_newline",
        "adapter.image_sep": "vit.perceive.image_sep",
    },
}

# Combine base mappings with overrides
# KEY_MAPS = {model: {**BASE_KEY_MAP, **overrides} for model, overrides in MODEL_OVERRIDES.items()}
KEY_MAPS = {
    model: recursive_update_dict(deepcopy(BASE_KEY_MAP), overrides, {}) for model, overrides in MODEL_OVERRIDES.items()
}

# Layer norm type
LN_TABLE = defaultdict(
    lambda: "rms",
    {
        "PhiForCausalLM": "standard",
        "GPT2LMHeadModel": "standard",
        "XLMRobertaXLForMaskedLM": "standard",
        "Gemma2ForCausalLM": "gemma-rms",
        "M2M100ForConditionalGeneration": "standard",
        "Gemma3ForConditionalGeneration": "gemma-rms",
    },
)

# Activation type (gated-silu also enables the ffn gate)
ACT_TABLE = defaultdict(
    lambda: "gated-silu",
    {
        "PhiForCausalLM": "gelu",
        "GPT2LMHeadModel": "gelu",
        "XLMRobertaXLForMaskedLM": "gelu",
        "Gemma2ForCausalLM": "gated-gelu",
        "Gemma3ForConditionalGeneration": "gated-gelu-tanh",
        "M2M100ForConditionalGeneration": "relu",
    },
)

# Not used anymore since Gemma3 PR, not sure if needed
# VISION_ACT_TABLE = defaultdict(
#     lambda: "gated-silu",
#     {
#         "Mistral3ForConditionalGeneration": "gated-gelu",
#     },
# )

# Eole config class
ARCH_TABLE = defaultdict(
    lambda: TransformerLMModelConfig,
    {
        "XLMRobertaXLForMaskedLM": TransformerEncoderModelConfig,
        "LlavaForConditionalGeneration": VisionTransformerLMModelConfig,
        "Mistral3ForConditionalGeneration": VisionTransformerLMModelConfig,
        "Gemma3ForConditionalGeneration": VisionTransformerLMModelConfig,
        "M2M100ForConditionalGeneration": TransformerModelConfig,
        "DeepseekOCRForCausalLM": VisionTransformerLMModelConfig,
        "HunYuanVLForConditionalGeneration": VisionTransformerLMModelConfig,
    },
)

# Default tokenization transform
TOK_TABLE = defaultdict(lambda: "huggingface_tokenize")
